#include "common.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/************************************
 *  Useful structures and handlers  *
 ************************************/


// Vector of longs
//-----------------


LongVec* longVecNew(long size) 
{
    LongVec* vec=NULL;
    void* ptr=NULL;

    if ((vec = calloc(1,sizeof(LongVec))) == NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;
    }

    if (size == -1)
	size = LONGVEC_INITSIZE;

    if ((ptr = calloc(size,sizeof(long))) == NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	free(vec);
	return NULL;
    }

    vec->size = size;
    vec->top = 0;
    vec->cont = ptr;

    return vec;
}

int longVecInit(LongVec* vec,long size)
{
    void* ptr=NULL;

    if (vec == NULL)
    {
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (size == -1)
	size = LONGVEC_INITSIZE;

    if ((ptr = calloc(size,sizeof(long))) == NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    vec->size = size;
    vec->top = 0;
    vec->cont = ptr;

    return 0;
}

int longVecEnd(LongVec* vec) 
{
    if (vec == NULL || vec->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    free(vec->cont);
    vec->cont = NULL;
    vec->size = 0;
    vec->top = 0;
    
    return 0;
}

int longVecDestroy(LongVec* vec)
{
    if (vec == NULL || vec->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    free(vec->cont);    
    free(vec);

    return 0;
}

int longVecCopy(LongVec *dst,LongVec *src)
{
    long *ptr=NULL;

    if (dst == NULL || dst->cont == NULL ||
	src == NULL || src->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (dst->size < src->size)
    {
	if ((ptr=(long*) realloc(dst->cont,src->size*sizeof(long)))==NULL)
	{
	    addError(CRITICAL,"Critical lack of memory in %s at line %d",
		     __FILE__,__LINE__);
	    return -1;
	}
	else
	{
	    dst->cont=ptr;
	    dst->size=src->size;
	}
    }

    memcpy(dst->cont,src->cont,src->size*sizeof(long));
    dst->top=src->top;

    return 0;
}

int longVecPushBack(LongVec* vec,long val)
{
    void* ptr=NULL;

    if (vec == NULL || vec->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (vec->top >= vec->size) 
    {
	if ((ptr = realloc(vec->cont, vec->size * 
			   2 * sizeof(long))) == NULL) 
	{
	    addError(CRITICAL,"Critical lack of memory in %s at line %d",
		     __FILE__,__LINE__);
	    return -1;
	}

	vec->size *= 2;
	vec->cont = ptr;
    }

    vec->cont[vec->top++] = val;

    return 0;
}

int longVecPopBack(LongVec* vec,long* val)
{
    if (vec == NULL || vec->cont == NULL || val == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (vec->top == 0) 
    { 
	addError(CRITICAL,"Pop on empty stack in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    *val = vec->cont[vec->top - 1];
    vec->top--;

    return 0;
}

int longVecResize(LongVec* vec,long newsize)
{
    void* ptr=NULL;

    if (vec == NULL || vec->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if ((ptr = realloc(vec->cont, newsize*sizeof(long))) == NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    vec->cont = ptr;
    memset(ptr+ (vec->size * sizeof(long)), 0, (newsize - vec->size)*sizeof(long));
    vec->size = newsize;

    return 0;
}

int longVecGet(LongVec* vec,long index, long* val)
{
    if (vec == NULL || vec->cont == NULL || val == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (index < 0)
    {
	addError(CRITICAL,"Bad argument (index < 0) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (index >= vec->size)
    {
	addError(CRITICAL,"Bad argument (wrong index) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    *val = vec->cont[index];

    return 0;

}

int longVecSet(LongVec* vec,long index,long val)
{
    if (vec == NULL || vec->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (index < 0)
    {
	addError(CRITICAL,"Bad argument (index < 0) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    if (index >= vec->size)
    {
	if (longVecResize(vec,max(vec->size * 2,index+1))<0)
	{
	    addError(CRITICAL,"Unable to resize long vector in %s at line %d",
		     __FILE__,__LINE__);
	    return -1;
	}
    }

    vec->cont[index] = val;
    vec->top = max(vec->top, index+1);

    return 0;
}


// List of longs
//---------------

int longListInsert(LongList *list,long index,long val)
{
    if (list == NULL || list->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (index < 0)
    {
	addError(CRITICAL,"Bad argument (index < 0) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if ((list->top >= list->size) || (index >= list->size))
    {
	if (longListResize(list,max(list->size * 2,index+1))<0)
	{
	    addError(CRITICAL,"Unable to resize long vector in %s at line %d",
		     __FILE__,__LINE__);
	    return -1;
	}
    }

    if (index < list->top)
    {
	memmove(list->cont+index+1,list->cont+index,(list->top-index) * sizeof(long));
    }
    list->cont[index]=val;	
    list->top++;
    
    return 0;
}

int longListRemove(LongList *list,long index)
{
    
    if (list == NULL || list->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (index < 0)
    {
	
	addError(CRITICAL,"Bad argument (index < 0) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    

    if (index>=list->top)
    {
	addError(WARNING,"Removing inexistent list element in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    else
    {
	memmove(list->cont+index,list->cont+index+1,(list->top-index-1) * sizeof(long));
	list->top--;
	return 0;
    }
}

int longCompare(const void *a,const void *b)
{
    if ((*(long*)a)<(*(long*)b))
	return -1;
    else if ((*(long*)a)>(*(long*)b))
	return 1;
    else
	return 0;
}

int longListMerge(LongList *la,LongList *lb,LongList *dest)
{
    int i=0,j=0;
    LongList tmpList;

    if (la == NULL || la->cont == NULL || 
	lb == NULL || lb->cont == NULL ||
	dest == NULL || dest->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    qsort(la->cont,la->top,sizeof(long),&longCompare);
    qsort(lb->cont,lb->top,sizeof(long),&longCompare); 
    
    if (longListInit(&tmpList,la->top+lb->top)<0)
    {
	addError(CRITICAL,"Unable to initialize temporary list of longs in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    while ((i<la->top) || (j<lb->top))
    {
	if ((j>=lb->top) || (i<la->top && la->cont[i]<lb->cont[j]))
	{
	    longListPushBack(&tmpList,la->cont[i]);
	    i++;	    
	}
	else if ((i>=la->top) || (j<lb->top && la->cont[i]>lb->cont[j]))
	{
	    longListPushBack(&tmpList,lb->cont[j]);
	    j++;	    
	}
	else if (la->cont[i]==lb->cont[j])
	{
	    longListPushBack(&tmpList,la->cont[i]);
	    i++;
	    j++;
	}
	else
	{
	    addError(CRITICAL,"Internal error in %s at line %d",
		     __FILE__,__LINE__);
	    longListEnd(&tmpList);
	    return -1;
	}
    }

    if (longListCopy(dest,&tmpList)<0)
    {
	addError(CRITICAL,"Unable to create resulting merged list of longs in %s at line %d",
		 __FILE__,__LINE__);
	longListEnd(&tmpList);
	return -1;
    }
    
    longListEnd(&tmpList);
    
    return 0;
}

int longListSort(LongList *list) {

    if (list == NULL || list->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    qsort(list->cont,list->top,sizeof(long),&longCompare);

    return 0;
}

// Vector of doubles
//-------------------


DoubleVec* dblVecNew(long size) 
{
    DoubleVec* vec=NULL;
    void* ptr=NULL;

    if ((vec = calloc(1,sizeof(DoubleVec))) == NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;
    }

    if (size == -1)
	size = DBLVEC_INITSIZE;

    if ((ptr = calloc(size,sizeof(double))) == NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	free(vec);
	return NULL;
    }

    vec->size = size;
    vec->top = 0;
    vec->cont = ptr;

    return vec;
}

int dblVecInit(DoubleVec* vec,long size)
{
    void* ptr=NULL;

    if (vec == NULL)
    {
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (size == -1)
	size = DBLVEC_INITSIZE;

    if ((ptr = calloc(size,sizeof(double))) == NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    vec->size = size;
    vec->top = 0;
    vec->cont = ptr;

    return 0;
}

int dblVecEnd(DoubleVec* vec) 
{
    if (vec == NULL || vec->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    free(vec->cont);
    vec->cont = NULL;
    vec->size = 0;
    vec->top = 0;
    
    return 0;
}

int dblVecDestroy(DoubleVec* vec)
{
    if (vec == NULL || vec->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    free(vec->cont);    
    free(vec);

    return 0;
}

int dblVecCopy(DoubleVec *dst,DoubleVec *src)
{
    double *ptr=NULL;

    if (dst == NULL || dst->cont == NULL ||
	src == NULL || src->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (dst->size < src->size)
    {
	if ((ptr=realloc(dst->cont,src->size*sizeof(double)))==NULL)
	{
	    addError(CRITICAL,"Critical lack of memory in %s at line %d",
		     __FILE__,__LINE__);
	    return -1;
	}
	else
	{
	    dst->cont=ptr;
	    dst->size=src->size;
	}
    }

    memcpy(dst->cont,src->cont,src->size*sizeof(double));
    dst->top=src->top;

    return 0;
}

int dblVecPushBack(DoubleVec *vec,double val)
{
    void* ptr=NULL;

    if (vec == NULL || vec->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (vec->top >= vec->size) 
    {
	if ((ptr = realloc(vec->cont, vec->size * 
			   2 * sizeof(double))) == NULL) 
	{
	    addError(CRITICAL,"Critical lack of memory in %s at line %d",
		     __FILE__,__LINE__);
	    return -1;
	}

	vec->size *= 2;
	vec->cont = ptr;
    }

    vec->cont[vec->top++] = val;

    return 0;
}

int dblVecPopBack(DoubleVec *vec,double *val)
{
    if (vec == NULL || vec->cont == NULL || val == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (vec->top == 0) 
    { 
	addError(CRITICAL,"Pop on empty stack in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    *val = vec->cont[vec->top - 1];
    vec->top--;

    return 0;
}


int dblVecResize(DoubleVec* vec,long newsize)
{
    void* ptr=NULL;

    if (vec == NULL || vec->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if ((ptr = realloc(vec->cont, newsize*sizeof(double))) == NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    vec->cont = ptr;
    memset(ptr+ (vec->size * sizeof(double)), 0, (newsize - vec->size)*sizeof(double));
    vec->size = newsize;

    return 0;
}

int dblVecGet(DoubleVec* vec,long index, double* val)
{
    if (vec == NULL || vec->cont == NULL || val == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (index < 0)
    {
	addError(CRITICAL,"Bad argument (index < 0) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (index >= vec->size)
    {
	addError(CRITICAL,"Bad argument (wrong index) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    *val = vec->cont[index];

    return 0;

}

int dblVecSet(DoubleVec* vec,long index,double val)
{
    if (vec == NULL || vec->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (index < 0)
    {
	addError(CRITICAL,"Bad argument (index < 0) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    if (index >= vec->size)
    {
	if (dblVecResize(vec,max(vec->size * 2,index+1))<0)
	{
	    addError(CRITICAL,"Unable to resize double vector in %s at line %d",
		     __FILE__,__LINE__);
	    return -1;
	}
    }

    vec->cont[index] = val;
    vec->top = max(vec->top, index+1);

    return 0;
}



/******************************
 *   Allocators redifinition  *
 ******************************/


#undef malloc
#undef realloc
#undef free
#undef calloc

long allocatedMemory=0;

void *mymalloc(size_t sz)
{
    void *ptr;

    if ((ptr=malloc(sz+sizeof(long)))==NULL)
	return NULL;
    
    allocatedMemory+=sz;
    *((long*)ptr)=sz;
    
    return (ptr+sizeof(long));
}

void *myrealloc(void *ptr,size_t sz)
{
    void *retptr;

    if ((retptr=realloc(ptr-sizeof(long),sz+sizeof(long)))==NULL)
	return NULL;

    allocatedMemory+=sz-(*((long*)(retptr)));
    *((long*)retptr)=sz;

    return (retptr+sizeof(long));
}

void myfree(void *ptr)
{
    if (*((long*)(ptr-sizeof(long))) == -1)
	fprintf(stderr, "Warning already freed\n");

    allocatedMemory-=*((long*)(ptr-sizeof(long)));
    *((long*)(ptr-sizeof(long)))=-1;
    free(ptr-sizeof(long));
}

void *mycalloc(size_t nmemb,size_t sz)
{
    void *ptr;

    if ((ptr=mymalloc(nmemb*sz))==NULL)
	return NULL;

    memset(ptr,0,nmemb*sz);

    return ptr;
}

