#include "computation/computation_api.h"
#include <string.h>


/*******************************************
 * Communication structures                *
 *******************************************/

// Label switched path Request representation
//--------------------------------------------

LSPRequest* lspRequestNew()
{
    LSPRequest* req;
    
    if ((req=calloc(1,sizeof(LSPRequest)))==NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;	
    }
    
    if (longListInit(&(req->forbidLinks),-1)<0)
    {
	free(req);
	addError(CRITICAL,"Unable to create label switched path request in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;
    }

    if (longListInit(&(req->path),-1)<0)
    {
	longListEnd(&(req->forbidLinks));
	free(req);
	addError(CRITICAL,"Unable to create label switched path request in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;
    }
    
    memset(req->bw, 0, NB_OA * sizeof(double));
    
    req->rerouteInfo.id = -1;

    return req; 
}

int lspRequestInit(LSPRequest *req)
{
    if (req == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }  

    memset(req,0,sizeof(LSPRequest));
    
    if (longListInit(&(req->forbidLinks),-1)<0)
    {
	addError(CRITICAL,"Unable to create label switched path request in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (longListInit(&(req->path),-1)<0)
    {
	longListEnd(&(req->forbidLinks));
	addError(CRITICAL,"Unable to create label switched path request in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    memset(req->bw, 0, NB_OA * sizeof(double));

    req->rerouteInfo.id = -1;

    return 0;
}

int lspRequestDestroy(LSPRequest *req)
{
    if (req == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }    

    longListEnd(&(req->forbidLinks));
    longListEnd(&(req->path));
    free(req);

    return 0;
}

int lspRequestEnd(LSPRequest *req)
{
    if (req == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }   

    longListEnd(&(req->forbidLinks));
    longListEnd(&(req->path));

    return 0;
}

int lspRequestCopy(LSPRequest *dst,LSPRequest *src)
{
    int ret=0;
    
    if (dst == NULL || src==NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    dst->id=src->id;
    dst->primID=src->primID;
    dst->precedence=src->precedence;
    dst->type=src->type;
    memmove(&(dst->rerouteInfo), &(src->rerouteInfo), sizeof(LSPrerouteInfo));
    ANDERROR(ret,longListCopy(&(dst->forbidLinks),&(src->forbidLinks)));
    ANDERROR(ret,longListCopy(&(dst->path),&(src->path)));

    if (ret<0)
    {
	addError(CRITICAL,"Label switched path request copy uncomplete in %s at line %d",
		 __FILE__,__LINE__);
    }
    
    return ret;
}


// API for the LSPRequestList
//----------------------------

int lspRequestListInit(LSPRequestList* reqList, long size)
{
    long i;

    if (reqList == NULL) 
    {
	addError(CRITICAL,"LSPRequestList == NULL in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    size = (size<=0?LSPREQLIST_INITSIZE:size);

    if ((reqList->cont = calloc(size, sizeof(LSPRequest))) == NULL)
    {
	addError(CRITICAL,"Impossible to allocate memory for LSPRequestList in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    for (i=0; i<size; i++)
    {
	if (lspRequestInit(&(reqList->cont[i])) < 0) 
	{
	    addError(CRITICAL,"Error while initialisation of LSPRequest in %s at line %d",
		 __FILE__,__LINE__);
	    // clean up
	    for (i--;i>=0;i--) 
	    {
		lspRequestEnd(&(reqList->cont[i]));
	    }
	    free(reqList->cont);
	    return -1;
	}
    }

    reqList->size = size;

    return 0;
}

int lspRequestListEnd(LSPRequestList* reqList)
{
    long i;

    if (reqList == NULL) 
    {
	addError(CRITICAL,"LSPRequestList == NULL in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    for (i=0; i<reqList->size; i++)
    {
	lspRequestEnd(&(reqList->cont[i]));
    }

    free(reqList->cont);

    return 0;
}

int lspRequestListResize(LSPRequestList* reqList, long size)
{
    long i;

    if (reqList == NULL) 
    {
	addError(CRITICAL,"LSPRequestList == NULL in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (reqList->size < size) 
    {
	LSPRequest* ptr;

	if ((ptr = realloc(reqList->cont, size*sizeof(LSPRequest))) == NULL)
	{
	    addError(CRITICAL,"Impossible to allocate memory for LSPRequestList in %s at line %d",
		     __FILE__,__LINE__);
	    return -1;
	}

	reqList->cont = ptr;

	for (i=reqList->size; i<size; i++)
	{
	    if (lspRequestInit(&(reqList->cont[i])) < 0) 
	    {
		addError(CRITICAL,"Error while initialisation of LSPRequest in %s at line %d",
			 __FILE__,__LINE__);
		// clean up
		for (i--;i>=reqList->size;i--) 
		{
		    lspRequestEnd(&(reqList->cont[i]));
		}
		return -1;
	    }
	}
	reqList->size = size;
    }
    else 
    {
	for (i=reqList->size-1; i>=size; i--)
	{
	    lspRequestEnd(&(reqList->cont[i])); 
	}
	reqList->size = size;
    }

    return 0;
}

long lspRequestListSize(LSPRequestList* reqList)
{
    if (reqList == NULL) 
    {
	addError(CRITICAL,"LSPRequestList == NULL in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    return reqList->size;
}

LSPRequest* lspRequestListGet(LSPRequestList* reqList, long index)
{
    if (reqList == NULL) 
    {
	addError(CRITICAL,"LSPRequestList == NULL in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;
    }

    if (index < 0 || index >= reqList->size) 
    {
	addError(CRITICAL,"Out of bound index in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;
    }

    return &(reqList->cont[index]);
}

int lspRequestListSet(LSPRequestList* reqList, LSPRequest* req, long index)
{
    addError(CRITICAL,"Trying to call an undefined function in %s at line %d",
	     __FILE__,__LINE__);

    return 0;
}
