#ifndef COMPUTATION_ST_H
#define COMPUTATION_ST_H

#include "common/common.h"
#include "error/error.h"
#include "database/database_st.h"
#include "computation/backup/backup_st.h"

/*******************************************
 * Communication structures                *
 *******************************************/

/** 
 * Rerouting Information structure.
 * Used to support soft preemption. When a LSP is preempted, we have two choices.  1. Tear down this LSP immediately, this is hard preemption.  2. Notice the entity responsible for this LSP (e.g. the ingress in a decentralized mode) so that it can reestablish another LSP before the preempted one is being torn down.  This is soft preemption.  When soft preemption is used, when the computation of the new LSP (meant for replacing the soon preempted one) occurs, the computation algorithm must take into account the fact that the ressources of the preempted one can be used.  But it is also interesting to take into account the link where the preemption occured, because it's certainly a link that must no more be used.  In a decentralized approach, there's a good probability that the topology representation that the ingress has is not up-to-date when computing the rerouting.  So, this is at least an interesting information to give to the computation algorithm. 
 */
typedef struct LSPrerouteInfo_
{
    long id; /**< id of the preempted lsp.  That is the id of the lsp of which this lsp is the rerouting. */
    long src; /**< the source of the link where preemption occurs */
    long dst; /**< the destination of the link where preemption occurs */
} LSPrerouteInfo;

/**
 * LSP Request Structure.
 * Label Switched Path request representation, used by #computePrimaryPath
 */
typedef struct LSPRequest_
{
    long id; /**< id of the LSP */
    long primID; /**< id of the primary LSP if the LSP is a backup */
    LSPrerouteInfo rerouteInfo;
    int precedence; /**< preemption (or priority) level */
    double bw[NB_OA]; /**< LSP bandwidth.  Note that to be Diff-Serv Aware TE compliant, this field should be different from 0 only for one OA, because multiple OAs are not allowed on the same LSP */
    LongList forbidLinks; /**< list of the link colors that can't be used */
    LongList path; /**< path of the LSP. When using LSPRequest as an argument to #computePrimaryPath, this list is filled with (src, -1, dst).  When #computePrimaryPath returns, this list contains the complete computed path */
    DBLSPType type; /**< can be PRIM, GLOBAL_BACK or LOCAL_BACK */
} LSPRequest;


// Label switched path Request list representation
//-------------------------------------------------

typedef struct LSPRequestList_ 
{
    LSPRequest *cont;
    long size;
} LSPRequestList;


#endif
