#include "computation/computation_api.h"
#include "database/database_api.h"
#include "common/common.h"
#include "common/setup.h"
#include "predicate/predicate.h"
#include "primaryPath_api.h"
#include "primaryPath_util.h"
#include <stdio.h>
#include <string.h>
#include <math.h>

#if defined LINUX && defined TIMING
#include <sys/time.h>
#endif


/**
 * Primary LSP computation function
 * @param dataBase the general database containing topology
 * @param req the request containing information about the lsp to be computed 
 */
int computePrimaryPath(DataBase *dataBase, LSPRequest *req)
{
    BKTopology topo;
    long src;
    long dst;
    BKNode *tmpNode;

#if defined LINUX && defined TIME4
    struct timezone tz;
    struct timeval  t1,t2;
#endif

    if (dataBase == NULL || req==NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

#if defined LINUX && defined TIMING && defined TIME4
    gettimeofday(&t1, &tz);
#endif

    if ((src=getRequestSrc(req))<0)
    {
	addError(CRITICAL,"Unable to get requested source in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    } 

#if defined EBK
    fprintf(stderr,"*****************\n");
    fprintf(stderr,"* COMPUTING LSP *\n");
    fprintf(stderr,"* ID = %ld \n",req->id);
    fprintf(stderr,"*****************\n");
#endif

#if defined EBK
    fprintf(stderr,"init/fillTopo: Creating pruned topology\n");
#endif
    
    if (initTopo(&topo,-1)<0)
    {
	addError(CRITICAL,"Unable to initialize the topology structure in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    if (fillTopo(dataBase,req,&topo)<0)
    {
	addError(CRITICAL,"Unable to build topology in %s at line %d",
		 __FILE__,__LINE__);
	endTopo(&topo);
	return -1;
    }
    //printTopo(&topo);

#if defined EBK
    fprintf(stderr,"Calling BellMan Kalaba algo\n");
#endif


    if (bellmanKalaba(&topo,req)<0)
    {
	addError(CRITICAL,"Bellman-Kalaba failure in %s at line %d",
		 __FILE__,__LINE__);
	endTopo(&topo);
	return -1;
    }

#if defined EBK
    fprintf(stderr,"Building path from BK topology\n");
#endif 


    if (updateRequest(&topo,req)<0)
    {
	addError(CRITICAL,"Unable to update requested path in %s at line %d",
		 __FILE__,__LINE__);
	endTopo(&topo);
	return -1;
    }	

// searching the last hop and then getting its cost

  
    if ((dst=getRequestDst(req))<0)
    {
	addError(CRITICAL,"Unable to get requested source in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    } 
    
    
    if ((tmpNode=bkNodeVecGet(&topo.nodeVec,topo.nodeInd.cont[dst]))==NULL)
    {
	addError(CRITICAL,"Unable to get egress node in %s at line %d",
		 __FILE__,__LINE__);

	return -1;
    }

    //fprintf(stderr,"Valeur du cot: %f\n", tmpNode->info.cost);
    COST+=tmpNode->info.cost;



#if defined LINUX && defined TIMING && defined TIME4
    gettimeofday(&t2, &tz);
    fprintf(stderr, "Time for calculation of primary path : %f ms\n", (t2.tv_sec - t1.tv_sec) * 1000 + 
	    (t2.tv_usec - t1.tv_usec) / 1000.0);
#endif

    endTopo(&topo);

    return 0;
}


// Vector of BKConnect
//---------------------

BKConnectVec* bkConnectVecNew(long size)
{
    BKConnectVec *vec=NULL;
    BKConnect *ptr=NULL;

    if ((vec = calloc(1,sizeof(BKConnectVec))) == NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;
    }

    if (size == -1)
	size = BKCONNECTVEC_INITSIZE;

    if ((ptr = (BKConnect*) calloc(size,sizeof(BKConnect))) == NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	free(vec);
	return NULL;
    }

    vec->size = size;
    vec->top = 0;
    vec->cont = ptr;

    return vec;
}

int bkConnectVecInit(BKConnectVec *vec,long size)
{
    BKConnect *ptr=NULL;

    if (vec == NULL)
    {
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (size == -1)
	size = BKCONNECTVEC_INITSIZE;

    if ((ptr = (BKConnect*) calloc(size,sizeof(BKConnect))) == NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    vec->size = size;
    vec->top = 0;
    vec->cont = ptr;

    return 0;
}

int bkConnectVecEnd(BKConnectVec *vec)
{
    if (vec == NULL || vec->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    free(vec->cont);
    vec->cont = NULL;
    vec->size = 0;
    vec->top = 0;
    
    return 0;
}

int bkConnectVecDestroy(BKConnectVec *vec)
{
    if (vec == NULL || vec->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    free(vec->cont);    
    free(vec);

    return 0;
}

int bkConnectVecCopy(BKConnectVec *dst,BKConnectVec *src)
{
    BKConnect *ptr=NULL;

    if (dst == NULL || dst->cont == NULL ||
	src == NULL || src->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    if (dst->size < src->size)
    {
	if ((ptr=(BKConnect*) realloc(dst->cont,src->size*sizeof(BKConnect)))==NULL)
	{
	    addError(CRITICAL,"Critical lack of memory in %s at line %d",
		     __FILE__,__LINE__);
	    return -1;
	}
	else
	{
	    dst->cont=ptr;
	    dst->size=src->size;
	}
    }
    
    memcpy(dst->cont,src->cont,src->size*sizeof(BKConnect));
    dst->top=src->top;

    return 0;
}

int bkConnectVecPushBack(BKConnectVec *vec,BKConnect *val)
{
    void* ptr=NULL;

    if (vec == NULL || vec->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (vec->top >= vec->size) 
    {
	if ((ptr = realloc(vec->cont, vec->size * 
			   2 * sizeof(BKConnect))) == NULL) 
	{
	    addError(CRITICAL,"Critical lack of memory in %s at line %d",
		     __FILE__,__LINE__);
	    return -1;
	}

	vec->size *= 2;
	vec->cont = ptr;
    }

    vec->cont[vec->top].neighbId = val->neighbId;
    vec->cont[vec->top++].linkState = val->linkState; // pointeur directement sur la DB! 

    return 0;
}

int bkConnectVecPopBack(BKConnectVec *vec,BKConnect *val)
{
    if (vec == NULL || vec->cont == NULL || val == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (vec->top == 0) 
    { 
	addError(CRITICAL,"Pop on empty stack in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    val->neighbId = vec->cont[vec->top - 1].neighbId;
    val->linkState = vec->cont[vec->top--].linkState;

    return 0;
}

int bkConnectVecResize(BKConnectVec *vec,long newsize)
{
    void* ptr=NULL;

    if (vec == NULL || vec->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if ((ptr = realloc(vec->cont, newsize*sizeof(BKConnect))) == NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    vec->cont = ptr;
    memset(ptr+ (vec->size * sizeof(BKConnect)), 0, (newsize - vec->size)*sizeof(BKConnect));
    vec->size = newsize;

    return 0;
}

int bkConnectVecGet(BKConnectVec *vec,long index,BKConnect *val)
{
    if (vec == NULL || vec->cont == NULL || val == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (index < 0)
    {
	addError(CRITICAL,"Bad argument (index < 0) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (index >= vec->size)
    {
	addError(CRITICAL,"Bad argument (wrong index) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    vec->cont[index].neighbId = val->neighbId;
    vec->cont[index].linkState = val->linkState; // pointeur directement sur la DB! 
    
    return 0;
}

int bkConnectVecSet(BKConnectVec *vec,long index,BKConnect *val)
{
    if (vec == NULL || vec->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (index < 0)
    {
	addError(CRITICAL,"Bad argument (index < 0) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    if (index >= vec->size)
    {
	if (bkConnectVecResize(vec,max(vec->size * 2,index+1))<0)
	{
	    addError(CRITICAL,"Unable to resize vector in %s at line %d",
		     __FILE__,__LINE__);
	    return -1;
	}
    }

    vec->cont[index].neighbId = val->neighbId;
    vec->cont[index].linkState = val->linkState; // pointeur directement sur la DB! 
    vec->top=max(vec->top,index+1);

    return 0;
}



// Vector of BKNode
//------------------

BKNodeVec* bkNodeVecNew(long size)
{
    BKNodeVec* vec=NULL;
    BKNode* ptr=NULL;
    long i,j;

    if ((vec = calloc(1,sizeof(BKNodeVec))) == NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;
    }

    if (size == -1)
	size = BKNODEVEC_INITSIZE;

    if ((ptr = calloc(size,sizeof(BKNode))) == NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	free(vec);
	return NULL;
    }

    for (i=0;i<size;i++)
    {
	if (bkConnectVecInit(&ptr[i].inNeighb,-1)<0)
	{
	    for (j=i-1;j>=0;j--)
	    {
		bkConnectVecEnd(&ptr[j].inNeighb);
		bkConnectVecEnd(&ptr[j].outNeighb);
	    }
	    addError(CRITICAL,"Unable to initialize structure in %s at line %d",
		     __FILE__,__LINE__);
	    free(vec);
	    free(ptr);
	    return NULL;
	}
	else if (bkConnectVecInit(&ptr[i].outNeighb,-1)<0)
	{
	    bkConnectVecEnd(&ptr[i].inNeighb);
	    for (j=i-1;j>=0;j--)
	    {
		bkConnectVecEnd(&ptr[j].inNeighb);
		bkConnectVecEnd(&ptr[j].outNeighb);
	    }
	    addError(CRITICAL,"Unable to initialize structure in %s at line %d",
		     __FILE__,__LINE__);
	    free(vec);
	    free(ptr);
	    return NULL;
	}
    }

    vec->size = size;
    vec->top = 0;
    vec->cont = ptr;

    return vec;
}

int bkNodeVecInit(BKNodeVec *vec,long size)
{
    BKNode* ptr=NULL;
    long i,j;

    if (vec == NULL)
    {
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (size == -1)
	size = BKNODEVEC_INITSIZE;

    if ((ptr = calloc(size,sizeof(BKNode))) == NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    for (i=0;i<size;i++)
    {
	if (bkConnectVecInit(&ptr[i].inNeighb,-1)<0)
	{
	    for (j=i-1;j>=0;j--)
	    {
		bkConnectVecEnd(&ptr[j].inNeighb);
		bkConnectVecEnd(&ptr[j].outNeighb);
	    }
	    addError(CRITICAL,"Unable to initialize structure in %s at line %d",
		     __FILE__,__LINE__);
	    free(ptr);
	    return -1;
	}
	else if (bkConnectVecInit(&ptr[i].outNeighb,-1)<0)
	{
	    bkConnectVecEnd(&ptr[i].inNeighb);
	    for (j=i-1;j>=0;j--)
	    {
		bkConnectVecEnd(&ptr[j].inNeighb);
		bkConnectVecEnd(&ptr[j].outNeighb);
	    }
	    addError(CRITICAL,"Unable to initialize structure in %s at line %d",
		     __FILE__,__LINE__);
	    free(ptr);
	    return -1;
	}
    }

    vec->size = size;
    vec->top = 0;
    vec->cont = ptr;

    return 0;
}

int bkNodeVecEnd(BKNodeVec *vec)
{
    long i;

    if (vec == NULL || vec->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    for (i=0;i<vec->size;i++)
    {
	bkConnectVecEnd(&vec->cont[i].inNeighb);
	bkConnectVecEnd(&vec->cont[i].outNeighb);
    }

    free(vec->cont);
    vec->cont = NULL;
    vec->size = 0;
    vec->top = 0;
    
    return 0;
}

int bkNodeVecDestroy(BKNodeVec *vec)
{
    long i;

    if (vec == NULL || vec->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    for (i=0;i<vec->size;i++)
    {
	bkConnectVecEnd(&vec->cont[i].inNeighb);
	bkConnectVecEnd(&vec->cont[i].outNeighb);
    }

    free(vec->cont);    
    free(vec);

    return 0;
}

int bkNodeVecPushBack(BKNodeVec *vec,BKNode *val)
{
    if (vec == NULL || vec->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (vec->top >= vec->size) 
    {
	if (bkNodeVecResize(vec,vec->size*2)<0)
	{
	    addError(CRITICAL,"Critical lack of memory in %s at line %d",
		     __FILE__,__LINE__);
	    return -1;
	}
    }

    if (bkConnectVecCopy(&vec->cont[vec->top].inNeighb,&val->inNeighb)<0)
    {
	addError(CRITICAL,"Unable to copy the list of neighbours in %s at line %d",
		     __FILE__,__LINE__);
	return -1;
    }
    if (bkConnectVecCopy(&vec->cont[vec->top].outNeighb,&val->outNeighb)<0)
    {
	addError(CRITICAL,"Unable to copy the list of neighbours in %s at line %d",
		     __FILE__,__LINE__);
	return -1;
    }
    vec->cont[vec->top].nodeId = val->nodeId;
    vec->cont[vec->top++].neighbInd = val->neighbInd;

    return 0;
}

int bkNodeVecPopBack(BKNodeVec *vec,BKNode *val)
{
    if (vec == NULL || vec->cont == NULL || val == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (vec->top == 0) 
    { 
	addError(CRITICAL,"Pop on empty stack in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (bkConnectVecCopy(&val->inNeighb,&vec->cont[vec->top-1].inNeighb)<0)
    {
	addError(CRITICAL,"Unable to copy the list of neighbours in %s at line %d",
		     __FILE__,__LINE__);
	return -1;
    }
    if (bkConnectVecCopy(&val->outNeighb,&vec->cont[vec->top-1].outNeighb)<0)
    {
	addError(CRITICAL,"Unable to copy the list of neighbours in %s at line %d",
		     __FILE__,__LINE__);
	return -1;
    }
    val->nodeId = vec->cont[vec->top-1].nodeId;
    val->neighbInd = vec->cont[vec->top--].neighbInd;

    return 0;
}

int bkNodeVecResize(BKNodeVec *vec,long newsize)
{
    void *ptr=NULL;
    long i;

    if (vec == NULL || vec->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if ((ptr = (BKNode*) realloc(vec->cont, newsize*sizeof(BKNode))) == NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    memset(ptr+(vec->size*sizeof(BKNode)) , 0, (newsize-vec->size)*sizeof(BKNode));
    vec->cont = ptr;    
    
    for (i=vec->size;i<newsize;i++)
    {
	if (bkConnectVecInit(&((BKNode*) ptr)[i].inNeighb,-1)<0 ||
	    bkConnectVecInit(&((BKNode*) ptr)[i].outNeighb,-1)<0)
	{
	    addError(CRITICAL,"Unable to initialize structure in %s at line %d",
		     __FILE__,__LINE__);
	    return -1;
	}
    }    
    vec->size = newsize;

    return 0;
}

BKNode *bkNodeVecGet(BKNodeVec *vec,long index)
{
    if (vec == NULL || vec->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;
    }

    if (index < 0)
    {
	addError(CRITICAL,"Bad argument (index < 0) in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;
    }

    if (index >= vec->size)
    {
	addError(CRITICAL,"Bad argument (wrong index) in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;
    }
    
    return vec->cont+index;
}

int bkNodeVecSet(BKNodeVec *vec,long index,BKNode *val)
{
    if (vec == NULL || vec->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (index < 0)
    {
	addError(CRITICAL,"Bad argument (index < 0) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    if (index >= vec->size)
    {
	if (bkNodeVecResize(vec,max(vec->size * 2,index+1))<0)
	{
	    addError(CRITICAL,"Unable to resize node vector in %s at line %d",
		     __FILE__,__LINE__);
	    return -1;
	}
    }
    
    if (bkConnectVecCopy(&vec->cont[index].inNeighb,&val->inNeighb)<0)
    {
	addError(CRITICAL,"Unable to copy the list of neighbours in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    if (bkConnectVecCopy(&vec->cont[index].outNeighb,&val->outNeighb)<0)
    {
	addError(CRITICAL,"Unable to copy the list of neighbours in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    vec->cont[index].nodeId = val->nodeId;
    vec->cont[index].neighbInd = val->neighbInd; 
    vec->top=max(vec->top,index+1);

    return 0;
}


// Auxiliary functions
//---------------------

int initTopo(BKTopology *topo, long size)
{
    if (topo == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    if (bkNodeVecInit(&topo->nodeVec,-1)<0)
    {
	addError(CRITICAL,"Unable to initialize node vector in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    if (longVecInit(&topo->nodeInd,size)<0)
    {
	addError(CRITICAL,"Unable to initialize long vector in %s at line %d",
		 __FILE__,__LINE__);
	bkNodeVecEnd(&topo->nodeVec);
	return -1;
    }    

    return 0;
}

int endTopo(BKTopology *topo)
{
    if (topo == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    bkNodeVecEnd(&topo->nodeVec);
    longVecEnd(&topo->nodeInd);

    return 0;
}


int fillTopo(DataBase *dataBase,LSPRequest *req,BKTopology *topo)
{
    LongList toDoNodes;
    int *activeFlags;
    LongList *tmpNeighb;
    long i,j,nd,src,size;
    BKConnect tmpConn;
    BKNode tmpNode,*nodePtr;

    
    if (dataBase == NULL || req==NULL || topo==NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
       
    if ((src=getRequestSrc(req))<0)
    {
	addError(CRITICAL,"Unable to get requested source in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    size=DBgetMaxNodeID(dataBase)+1;
    
    if (longListInit(&toDoNodes,size)<0)
    {
	addError(CRITICAL,"Unable to initialize the active nodes list in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if ((activeFlags = (int*) calloc(size,sizeof(long))) == NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	longListEnd(&toDoNodes);
	return -1;
    }

    memset(&tmpNode,0,sizeof(BKNode));
    if (bkConnectVecInit(&tmpNode.inNeighb,-1)<0)
    {
	addError(CRITICAL,"Unable to initialize the temporary node in %s at line %d",
		 __FILE__,__LINE__);
	longListEnd(&toDoNodes);
	free(activeFlags);
	return -1;
    }
    if (bkConnectVecInit(&tmpNode.outNeighb,-1)<0)
    {
	addError(CRITICAL,"Unable to initialize the temporary node in %s at line %d",
		 __FILE__,__LINE__);
	longListEnd(&toDoNodes);
	free(activeFlags);
	bkConnectVecEnd(&tmpNode.inNeighb);
	return -1;
    }
    
    if (longListPushBack(&toDoNodes,src)<0)
    {
	addError(CRITICAL,"Unable to push back on list of longs in %s at line %d",
		 __FILE__,__LINE__);
	longListEnd(&toDoNodes);
	free(activeFlags);
	bkConnectVecEnd(&tmpNode.inNeighb);
	bkConnectVecEnd(&tmpNode.outNeighb);
	return -1;
    } 
    activeFlags[src]=1;
    while (toDoNodes.top>0)
    {
	if (longListPopBack(&toDoNodes,&nd)<0)
	{
	    addError(CRITICAL,"Unable to pop back on list of longs in %s at line %d",
		     __FILE__,__LINE__);
	    longListEnd(&toDoNodes);
	    free(activeFlags);
	    bkConnectVecEnd(&tmpNode.inNeighb);
	    bkConnectVecEnd(&tmpNode.outNeighb);
	    return -1;
	}    
	
	tmpNode.inNeighb.top=0;
	if ((tmpNeighb=DBgetNodeInNeighb(dataBase,nd))==NULL)
	{
	    addError(CRITICAL,"Unable to get the list of neighbours in %s at line %d",
		     __FILE__,__LINE__);
	    longListEnd(&toDoNodes);
	    free(activeFlags);
	    bkConnectVecEnd(&tmpNode.inNeighb);
	    bkConnectVecEnd(&tmpNode.outNeighb);
	    return -1;
	}
	for (i=0;i<tmpNeighb->top;i++)
	{
	    if (activeFlags[tmpNeighb->cont[i]]==2)
	    {
		nodePtr=&(topo->nodeVec.cont[topo->nodeInd.cont[tmpNeighb->cont[i]]]);
		for (j=0;(j<nodePtr->outNeighb.top) && (nodePtr->outNeighb.cont[j].neighbId!=nd);j++);
		if (j<nodePtr->outNeighb.top)
		{
		    tmpConn.neighbId=tmpNeighb->cont[i];
		    tmpConn.linkState=nodePtr->outNeighb.cont[j].linkState;
		    memset(&tmpConn.info,0,sizeof(BKConnectInfo));
		    memcpy(tmpConn.info.gain,nodePtr->outNeighb.cont[j].info.gain,NB_OA*sizeof(double));
		    if (bkConnectVecPushBack(&tmpNode.inNeighb,&tmpConn)<0)
		    {
			addError(CRITICAL,"Unable to push back neighbour in %s at line %d",
				 __FILE__,__LINE__);
			longListEnd(&toDoNodes);
			free(activeFlags);
			bkConnectVecEnd(&tmpNode.inNeighb);
			bkConnectVecEnd(&tmpNode.outNeighb);
			return -1;
		    }	
		}
	    }
	    else
	    {
		tmpConn.neighbId=tmpNeighb->cont[i];
		tmpConn.linkState=DBgetLinkState(dataBase,tmpNeighb->cont[i],nd);
		memset(&tmpConn.info,0,sizeof(BKConnectInfo));
		if (isValidRequestLink(dataBase,tmpNeighb->cont[i],nd,
				       tmpConn.linkState,req,tmpConn.info.gain))
		{
		    if (bkConnectVecPushBack(&tmpNode.inNeighb,&tmpConn)<0)
		    {
			addError(CRITICAL,"Unable to push back neighbour in %s at line %d",
				 __FILE__,__LINE__);
			longListEnd(&toDoNodes);
			free(activeFlags);
			bkConnectVecEnd(&tmpNode.inNeighb);
			bkConnectVecEnd(&tmpNode.outNeighb);
			return -1;
		    }	
		}
	    }
	}
 
	tmpNode.outNeighb.top=0;
	if ((tmpNeighb=DBgetNodeOutNeighb(dataBase,nd))==NULL)
	{
	    addError(CRITICAL,"Unable to get the list of neighbours in %s at line %d",
		     __FILE__,__LINE__);
	    longListEnd(&toDoNodes);
	    free(activeFlags);
	    bkConnectVecEnd(&tmpNode.inNeighb);
	    bkConnectVecEnd(&tmpNode.outNeighb);
	    return -1;
	}
	for (i=0;i<tmpNeighb->top;i++)
	{
	    if (activeFlags[tmpNeighb->cont[i]]==2)
	    {
		nodePtr=&(topo->nodeVec.cont[topo->nodeInd.cont[tmpNeighb->cont[i]]]);
		for (j=0;(j<nodePtr->inNeighb.top) && (nodePtr->inNeighb.cont[j].neighbId!=nd);j++);
		if (j<nodePtr->inNeighb.top)
		{
		    tmpConn.neighbId=tmpNeighb->cont[i];
		    tmpConn.linkState=nodePtr->inNeighb.cont[j].linkState;
		    memset(&tmpConn.info,0,sizeof(BKConnectInfo));
		    memcpy(tmpConn.info.gain,nodePtr->inNeighb.cont[j].info.gain,NB_OA*sizeof(double));
		    if (bkConnectVecPushBack(&tmpNode.outNeighb,&tmpConn)<0)
		    {
			addError(CRITICAL,"Unable to push back neighbour in %s at line %d",
				 __FILE__,__LINE__);
			longListEnd(&toDoNodes);
			free(activeFlags);
			bkConnectVecEnd(&tmpNode.inNeighb);
			bkConnectVecEnd(&tmpNode.outNeighb);
			return -1;
		    }	
		}
	    }
	    else
	    {
		tmpConn.neighbId=tmpNeighb->cont[i];
		tmpConn.linkState=DBgetLinkState(dataBase,nd,tmpNeighb->cont[i]);
		memset(&tmpConn.info,0,sizeof(BKConnectInfo));
		if (isValidRequestLink(dataBase,nd,tmpNeighb->cont[i],
				       tmpConn.linkState,req,tmpConn.info.gain))
		{
		    if (bkConnectVecPushBack(&tmpNode.outNeighb,&tmpConn)<0)
		    {
			addError(CRITICAL,"Unable to push back neighbour in %s at line %d",
				 __FILE__,__LINE__);
			longListEnd(&toDoNodes);
			free(activeFlags);
			bkConnectVecEnd(&tmpNode.inNeighb);
			bkConnectVecEnd(&tmpNode.outNeighb);
			return -1;
		    }	
		}
	    }
	    if (activeFlags[tmpNeighb->cont[i]]==0)
	    {
		if (longListPushBack(&toDoNodes,tmpNeighb->cont[i])<0)
		{
		    addError(CRITICAL,"Unable to push back on list of longs in %s at line %d",
			     __FILE__,__LINE__);
		    longListEnd(&toDoNodes);
		    free(activeFlags);
		    return -1;
		}
		activeFlags[tmpNeighb->cont[i]]=1;
	    }
	}
	
	tmpNode.nodeId=nd;
	tmpNode.neighbInd=-1;
	if (bkNodeVecPushBack(&topo->nodeVec,&tmpNode)<0)
	{
	    addError(CRITICAL,"Unable to push back node in %s at line %d",
		     __FILE__,__LINE__);
	    longListEnd(&toDoNodes);
	    free(activeFlags);
	    bkConnectVecEnd(&tmpNode.inNeighb);
	    bkConnectVecEnd(&tmpNode.outNeighb);
	    return -1;
	}
	
	if (longVecSet(&topo->nodeInd,nd,(topo->nodeVec.top-1))<0)
	{
	    addError(CRITICAL,"Unable to set node index in %s at line %d",
		     __FILE__,__LINE__);
	    longListEnd(&toDoNodes);
	    free(activeFlags);
	    bkConnectVecEnd(&tmpNode.inNeighb);
	    bkConnectVecEnd(&tmpNode.outNeighb);
	    return -1;
	}

	activeFlags[nd]=2;
    }

    if (((topo->nbNodes=DBgetNbNodes(dataBase))<0)||
	((topo->nbLinks=DBgetNbLinks(dataBase))<0))
    {
	addError(CRITICAL,"Unable to get number of nodes and links in %s at line %d",
		 __FILE__,__LINE__);
	longListEnd(&toDoNodes);
	free(activeFlags);
	bkConnectVecEnd(&tmpNode.inNeighb);
	bkConnectVecEnd(&tmpNode.outNeighb);
	return -1;
    }

    longListEnd(&toDoNodes);
    free(activeFlags);
    bkConnectVecEnd(&tmpNode.inNeighb);
    bkConnectVecEnd(&tmpNode.outNeighb);

    return 0;
}

int printTopo(BKTopology *topo)
{
    BKNode *tmpNode;
    long i,j;

    for (i=0;i<topo->nodeInd.top;i++) 
    {
	tmpNode=bkNodeVecGet(&topo->nodeVec,topo->nodeInd.cont[i]);
	if (tmpNode!=NULL)
	{
	    if (i!=tmpNode->nodeId)
	    {
		addError(CRITICAL,"Topology unconsistancy in %s at line %d",
			 __FILE__,__LINE__);
		return -1;
	    }

	    printf("Node %ld\n--------\n",i);
	    printf("Incoming neighboors : \n");
	    
	    for (j=0; j<tmpNode->inNeighb.top; j++)
	    {
		printf("%ld ", tmpNode->inNeighb.cont[j].neighbId);
	    }
	    
	    printf("\nOutgoing neighboors : \n");
	    
	    for (j=0; j<tmpNode->outNeighb.top; j++)
	    {
		printf("%ld ", tmpNode->outNeighb.cont[j].neighbId);
	    }
	    printf("\n");

	    printf("Chosen Neighbour Index: %ld \n",tmpNode->neighbInd);	    
	    
	    printf("\n");
	}
    }

    return 0;
}

int getRequestSrc(LSPRequest *req)
{
    if (req==NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (req->path.top<2 || req->path.cont[0]<0 || 
	req->path.cont[req->path.top-1]<0)
    {
	addError(CRITICAL,"Bad requested path format in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    return req->path.cont[0];
}

int getRequestDst(LSPRequest *req)
{
    if (req==NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (req->path.top<2 || req->path.cont[0]<0 || 
	req->path.cont[req->path.top-1]<0)
    {
	addError(CRITICAL,"Bad requested path format in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    return req->path.cont[req->path.top-1];
}

int updateRequest(BKTopology *topo,LSPRequest *req)
{
    BKNode *tmpNode;
    long i,src,dst,nd;
    
    if (topo == NULL || req==NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if ((src=getRequestSrc(req))<0)
    {
	addError(CRITICAL,"Unable to get requested source in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if ((dst=getRequestDst(req))<0)
    {
	addError(CRITICAL,"Unable to get requested source in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }  

    req->path.top=0;    
    nd=dst;
    if ((tmpNode=bkNodeVecGet(&topo->nodeVec,topo->nodeInd.cont[nd]))==NULL)
    {
	addError(CRITICAL,"Undetermined error in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    while (nd!=src)
    {
	if (tmpNode->neighbInd < 0)
	{
	    addError(CRITICAL,"Destination unreachable in %s at line %d",
		     __FILE__,__LINE__);
	    return -1;
	}
	if (longListPushBack(&req->path,nd)<0)
	{
	    addError(CRITICAL,"Undetermined error in %s at line %d",
		     __FILE__,__LINE__);
	    return -1;
	}
	nd=tmpNode->inNeighb.cont[tmpNode->neighbInd].neighbId;
	if ((tmpNode=bkNodeVecGet(&topo->nodeVec,topo->nodeInd.cont[nd]))==NULL)
	{
	    addError(CRITICAL,"Undetermined error in %s at line %d",
		     __FILE__,__LINE__);
	    return -1;
	}
    }
    if (longListPushBack(&req->path,nd)<0)
    {
	addError(CRITICAL,"Undetermined error in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    for (i=0;i<req->path.top/2;i++)
    {
	nd=req->path.cont[i];
	req->path.cont[i]=req->path.cont[req->path.top-1-i];
	req->path.cont[req->path.top-1-i]=nd;
    }

    return 0;
}


int bellmanKalaba(BKTopology *topo,LSPRequest *req)
{
    LongList activeNodes;
    BKNode *tmpNode;
    bool done=FALSE;
    int *activeFlags;
    long src,i,j,k,nd,top,threshold,size,iter=0;
    double tmpCost;


    if (topo == NULL)
    {
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if ((src=getRequestSrc(req))<0)
    {
	addError(CRITICAL,"Unable to get requested source in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

 
    size=topo->nodeInd.top;

#if defined EBK
    fprintf(stderr,"Debug: Size value (nodeIndex.top) %ld\n", size);
#endif

    if (longListInit(&activeNodes,size)<0)
    {
	addError(CRITICAL,"Unable to initialize the active nodes list in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }


    if ((activeFlags = (int*) calloc(size,sizeof(long))) == NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	longListEnd(&activeNodes);
	return -1;
    }

    if (src>=size)
    {
	addError(CRITICAL,"Inexistent node in %s at line %d",
		 __FILE__,__LINE__);
	longListEnd(&activeNodes);
	free(activeFlags);
	return -1;
    }

#if defined EBK
    fprintf(stderr,"Initializing sum argument for the source node!, setting it to SUM(Load/Capacity)\n");
#endif 

    if (initScore(src,topo)<0)
    {
	addError(CRITICAL,"Unable to initialize scores in %s at line %d",
		 __FILE__,__LINE__);
	longListEnd(&activeNodes);
	free(activeFlags);
	return -1;
    }


    top=topo->nodeVec.cont[topo->nodeInd.cont[src]].outNeighb.top;

#if defined EBK
    fprintf(stderr,"Considering src node, here's the number of out neighbors: %ld\n", top);
#endif 

    for (i=0;i<top;i++)
    {
	nd=topo->nodeVec.cont[topo->nodeInd.cont[src]].outNeighb.cont[i].neighbId;
#if defined EBK
	fprintf(stderr," Consider neighbor: %ld\n", nd);
#endif 
	if (nd>=size)
	{
	    addError(CRITICAL,"Inexistent node in %s at line %d",
		     __FILE__,__LINE__);
	    longListEnd(&activeNodes);
	    free(activeFlags);
	    return -1;
	}

#if defined EBK
	fprintf(stderr,"Putting it in the activeNode List\n");
#endif
	if (longListPushBack(&activeNodes,nd)<0)
	{
	    addError(CRITICAL,"Undetermined error in %s at line %d",
		     __FILE__,__LINE__);
	    longListEnd(&activeNodes);
	    free(activeFlags);
	    return -1;
	}

	if ((tmpNode=bkNodeVecGet(&topo->nodeVec,topo->nodeInd.cont[nd]))==NULL)
	{
	    addError(CRITICAL,"Undetermined error in %s at line %d",
		     __FILE__,__LINE__);
	    longListEnd(&activeNodes);
	    free(activeFlags);
	    return -1;
	}

	for (k=0;(k<tmpNode->inNeighb.top) && (tmpNode->inNeighb.cont[k].neighbId!=src);k++);
	if (k>=tmpNode->inNeighb.top)
	{
	    addError(CRITICAL,"Topology unconsistancy in %s at line %d",
		     __FILE__,__LINE__);
	    longListEnd(&activeNodes);
	    free(activeFlags);
	    return -1;
	}
	

	tmpNode->info.cost=makeScore(topo,req,src,nd,&tmpNode->inNeighb.cont[k]);
	tmpNode->info.newCost=tmpNode->info.cost;
	tmpNode->neighbInd=k;
	tmpNode->info.newNeighbInd=tmpNode->neighbInd;

#if defined EBK
	fprintf(stderr,"We know that this node will necessarily be relaxed\n");
	fprintf(stderr,"Node Score(new): %f \n",tmpNode->info.newCost);
	fprintf(stderr,"Predecessor(new): %ld \n", tmpNode->info.newNeighbInd);
#endif
	updateNodeInfoOnElect(topo,req,src,nd,&tmpNode->inNeighb.cont[k]);
	activateNodeInfo(topo,nd);
	
#if defined EBK
	fprintf(stderr,"Status of node set to 1 (in process)\n");	
#endif
	activeFlags[nd]=1;
    }
#if defined EBK
	fprintf(stderr,"Status of source node set to 2 (treated)\n");
       
#endif   

    activeFlags[src]=2;


#if defined EBK
    fprintf(stderr,"Loop on boolean DONE\n");
#endif

    while (!done)
    {
	iter++;
#if defined EBK
	fprintf(stderr,"Iteration : %ld\n", iter-1);
	fprintf(stderr,"Setting DONE to TRUE --> abort\n");
#endif 	
	done=TRUE;

	threshold=activeNodes.top;

#if defined EBK
	fprintf(stderr,"Loop on the activeNodesList size: %ld\n", threshold);
#endif
	for (i=0;i<threshold;i++)
	{
	    top=topo->nodeVec.cont[topo->nodeInd.cont[activeNodes.cont[i]]].inNeighb.top;
#if defined EBK
	    fprintf(stderr,"Considering active node : %ld\n", activeNodes.cont[i]);
	    fprintf(stderr,"Loop on the number of incoming neighbours: %ld\n", top);
#endif

	    for (j=0;j<top;j++)
	    {
		nd=topo->nodeVec.cont[topo->nodeInd.cont[activeNodes.cont[i]]].inNeighb.cont[j].neighbId;
		if (nd>=size)
		{
		    addError(CRITICAL,"Inexistent node in %s at line %d",
			     __FILE__,__LINE__);
		    longListEnd(&activeNodes);
		    free(activeFlags);
		    return -1;
		}

#if defined EBK
		fprintf(stderr,"Considering incoming neighbour %ld with status: %ld\n", nd, activeFlags[nd]);
#endif
		if (activeFlags[nd]!=0 && noLoop(topo,nd,activeNodes.cont[i]))
		{

		    tmpCost=makeScore(topo,req,nd,activeNodes.cont[i],
				      &topo->nodeVec.cont[topo->nodeInd.cont[activeNodes.cont[i]]].inNeighb.cont[j]);

#if defined EBK
		    fprintf(stderr,"Status of node is not 0 and there's no loop when considering predecessors\n");
		    fprintf(stderr,"Potential cost for considered active node (link incoming neigh->active node): %f\n", tmpCost);
#endif
		    if (tmpCost-topo->nodeVec.cont[topo->nodeInd.cont[activeNodes.cont[i]]].info.cost<-1*DIGIT_PRECISION)
		    {

			done=FALSE;
			topo->nodeVec.cont[topo->nodeInd.cont[activeNodes.cont[i]]].info.newCost=tmpCost;
			topo->nodeVec.cont[topo->nodeInd.cont[activeNodes.cont[i]]].info.newNeighbInd=j;
			updateNodeInfoOnElect(topo,req,nd,activeNodes.cont[i],
					      &topo->nodeVec.cont[topo->nodeInd.cont[activeNodes.cont[i]]].inNeighb.cont[j]);

#if defined EBK
			fprintf(stderr,"It happens that this cost is lower than actual cost -> RELAXING (setting cost(new) to this value)\n");
			fprintf(stderr,"Cost(old): %f\n", topo->nodeVec.cont[topo->nodeInd.cont[activeNodes.cont[i]]].info.cost);
			fprintf(stderr,"Setting also (new) predecessor to the concerned incoming neighbor\n");
			fprintf(stderr,"Setting sum (new) value !\n");
			fprintf(stderr,"ALL OLD VALUES REMAIN AS THEY WERE\n");
#endif
		    }
		}
	    }


	    if (activeFlags[activeNodes.cont[i]]==1)
	    {
		top=topo->nodeVec.cont[topo->nodeInd.cont[activeNodes.cont[i]]].outNeighb.top;
#if defined EBK
		fprintf(stderr,"The concerned active node %ld has status (1), considering all outgoing neighbours (loop size: %ld)\n",activeNodes.cont[i],top);
#endif
		for (j=0;j<top;j++)
		{
		    nd=topo->nodeVec.cont[topo->nodeInd.cont[activeNodes.cont[i]]].outNeighb.cont[j].neighbId;

#if defined EBK
		    fprintf(stderr,"Considering out neighbor %ld\n", nd);
#endif
		    if (nd>=size)
		    {
			addError(CRITICAL,"Inexistent node in %s at line %d",
				 __FILE__,__LINE__);
			longListEnd(&activeNodes);
			free(activeFlags);
			return -1;
		    }

		    if (activeFlags[nd]==0)
		    {
#if defined EBK

			fprintf(stderr,"This neighbour has status 0 (not yet seen-> WE RELAX!), setting done to FALSE, adding it to active nodes list\n");
#endif
			done=FALSE;

			if (longListPushBack(&activeNodes,nd)<0)
			{
			    addError(CRITICAL,"Undetermined error in %s at line %d",
				     __FILE__,__LINE__);
			    longListEnd(&activeNodes);
			    free(activeFlags);
			    return -1;
			}

			if ((tmpNode=bkNodeVecGet(&topo->nodeVec,topo->nodeInd.cont[nd]))==NULL)
			{
			    addError(CRITICAL,"Undetermined error in %s at line %d",
				     __FILE__,__LINE__);
			    longListEnd(&activeNodes);
			    free(activeFlags);
			    return -1;
			}
			for (k=0;(k<tmpNode->inNeighb.top) &&
				 (tmpNode->inNeighb.cont[k].neighbId!=activeNodes.cont[i]);k++);
			if (k>=tmpNode->inNeighb.top)
			{
			    addError(CRITICAL,"Topology unconsistancy in %s at line %d",
				     __FILE__,__LINE__);
			    longListEnd(&activeNodes);
			    free(activeFlags);
			    return -1;
			}
			
			
			tmpNode->info.cost=makeScore(topo,req,activeNodes.cont[i],nd,&tmpNode->inNeighb.cont[k]);
			tmpNode->info.newCost=tmpNode->info.cost;
			tmpNode->neighbInd=k;
			tmpNode->info.newNeighbInd=tmpNode->neighbInd;
#if defined EBK
			fprintf(stderr,"We know that this node will necessarily be relaxed\n");
			fprintf(stderr,"Node Score(new): %f \n",tmpNode->info.newCost);
			fprintf(stderr,"Predecessor(new): %ld \n", tmpNode->info.newNeighbInd);
#endif
			updateNodeInfoOnElect(topo,req,activeNodes.cont[i],nd,&tmpNode->inNeighb.cont[k]);
			activateNodeInfo(topo,nd);
#if defined EBK
	fprintf(stderr,"Status of node set to 1 (in process)\n");	
#endif
			activeFlags[nd]=1;


		    }
		}
		activeFlags[activeNodes.cont[i]]=2;
#if defined EBK
		fprintf(stderr,"Active node %ld HAS BEEN TREATED (status set to 2)\n", activeNodes.cont[i]);
#endif
	    }
	    else if (activeFlags[activeNodes.cont[i]]==0)
	    {
		addError(CRITICAL,"Internal unconsistancy in %s at line %d",
			 __FILE__,__LINE__);
		longListEnd(&activeNodes);
		free(activeFlags);
		return -1;
	    }
	}
	
#if defined EBK
	fprintf(stderr,"Coming back in a loop for all prev-active nodes, threshold value: %ld\n", threshold);
	fprintf(stderr,"New values had not been set, we now set them\n");
	fprintf(stderr,"Think there may be an error here!\n");
#endif

	for (i=0;i<threshold;i++)
	{
	    if (activeFlags[activeNodes.cont[i]]==2)
	    {
#if defined EBK
		fprintf(stderr,"New cost value for concerned active node %ld: %f\n",activeNodes.cont[i],topo->nodeVec.cont[topo->nodeInd.cont[activeNodes.cont[i]]].info.cost );
#endif
		topo->nodeVec.cont[topo->nodeInd.cont[activeNodes.cont[i]]].info.cost=
		    topo->nodeVec.cont[topo->nodeInd.cont[activeNodes.cont[i]]].info.newCost;
#if defined EBK
		fprintf(stderr,"New neighbor value for concerned active node %ld: %ld\n",activeNodes.cont[i], topo->nodeVec.cont[topo->nodeInd.cont[activeNodes.cont[i]]].neighbInd);
#endif
		topo->nodeVec.cont[topo->nodeInd.cont[activeNodes.cont[i]]].neighbInd=
		    topo->nodeVec.cont[topo->nodeInd.cont[activeNodes.cont[i]]].info.newNeighbInd;

		activateNodeInfo(topo,activeNodes.cont[i]);
	    }
	}
    }

    longListEnd(&activeNodes);
    free(activeFlags);

#ifdef DEBUG
    printf("Bellman-Kalaba : %ld iterations \n",iter);
#endif

    return 0;
}

int initScore(long src,BKTopology *topo)
{
    bool process=FALSE;
    long i,j,k,l,top;
    double tmpSum;

    if (topo == NULL)
    {
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    for (i=0;i<NB_OA;i++)
    {
	if (damoteConfig.primaryComputationConfig.loadBal[i]!=0)
	{
	    process=TRUE;
	}
    }

    if (process)
    {
	for (k=0;k<NB_OA;k++)
	{
	    topo->nodeVec.cont[topo->nodeInd.cont[src]].info.sum[k]=0;
	}
	for (i=0;i<topo->nodeVec.top;i++)
	{
	    top=topo->nodeVec.cont[i].inNeighb.top;
	    for (j=0;j<top;j++)
	    {
		for (k=0;k<NB_OA;k++)
		{
		    tmpSum=0;
		    for (l=0;l<NB_PREEMPTION;l++)
		    {
			tmpSum+=topo->nodeVec.cont[i].inNeighb.cont[j].linkState->pbw[k][l];
		    }
		    topo->nodeVec.cont[topo->nodeInd.cont[src]].info.sum[k]+=tmpSum/
			topo->nodeVec.cont[i].inNeighb.cont[j].linkState->cap[k];
		}
	    }
	}
#if defined EBK
	fprintf(stderr,"Value of sum argument for src node (and OA 0): %f\n", topo->nodeVec.cont[topo->nodeInd.cont[src]].info.sum[0]);
#endif

    }

    return 0;
}


double makeScore(BKTopology *topo,LSPRequest *req,long src,long dst,BKConnect *connect)
{
    double score=0,totBW[NB_OA],newSum,rerouteScore=0;
    double capacity,load,utilization,phi,oldphi;
    long i,j;

    if (topo == NULL || connect == NULL)
    {
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return HUGE_VAL;
    }

    score=topo->nodeVec.cont[topo->nodeInd.cont[src]].info.cost;
#if defined EBK
    fprintf(stderr,"Printing src cost: %f\n", score);
#endif
  
    

    for (i=0;i<NB_OA;i++)
    {
	totBW[i]=0;
	for (j=0;j<NB_PREEMPTION;j++)
	{
	    totBW[i]+=connect->linkState->pbw[i][j];
	}
	
	
	capacity = connect->linkState->cap[i];
	
        load = totBW[i];
	
	utilization = load/capacity;
	
      
	
	if (utilization < ((float)1/3)) {
	    phi = load;
	} else if (utilization < ((float)2/3)) {
	    phi = (3 * load) - (((float)2/3) * capacity);
	} else if (utilization < (float)(9/10)) {
	    phi = (10 * load) - (((float)16/3) * capacity);
	} else if (utilization < 1) {
	    phi = (70 * load) - (((float)178/3) * capacity);
	} else if (utilization < ((float)11/10)) {
	    phi = (500 * load) - (((float)1468/3) * capacity);
	} else  {
	    phi = (5000 * load) - (((float)19468/3) * capacity);
	}
	
	oldphi = phi;
	
	load = totBW[i]+req->bw[i];
	
	if (utilization < ((float)1/3)) {
	    phi = load;
	} else if (utilization < ((float)2/3)) {
	    phi = (3 * load) - (((float)2/3) * capacity);
	} else if (utilization < (float)(9/10)) {
	    phi = (10 * load) - (((float)16/3) * capacity);
	} else if (utilization < 1) {
	    phi = (70 * load) - (((float)178/3) * capacity);
	} else if (utilization < ((float)11/10)) {
	    phi = (500 * load) - (((float)1468/3) * capacity);
	} else  {
	    phi = (5000 * load) - (((float)19468/3) * capacity);
	}
	
	
	score+=damoteConfig.primaryComputationConfig.Fortz*(phi-oldphi);
	
	if (damoteConfig.primaryComputationConfig.loadBal[i]!=0)
	{
	    score+=damoteConfig.primaryComputationConfig.loadBal[i]
		*((totBW[i]+req->bw[i])/connect->linkState->cap[i])
		*((totBW[i]+req->bw[i])/connect->linkState->cap[i]);
	    score-=damoteConfig.primaryComputationConfig.loadBal[i]
		*(totBW[i]/connect->linkState->cap[i])
		*(totBW[i]/connect->linkState->cap[i]);

	    newSum=topo->nodeVec.cont[topo->nodeInd.cont[src]].info.sum[i]+(req->bw[i]/connect->linkState->cap[i]);
	    if (__isinf(newSum))
	    {
		return HUGE_VAL;
	    }
	    
	    score+=damoteConfig.primaryComputationConfig.loadBal[i]
		*(-1/(double)topo->nbLinks)*newSum*newSum;
	    score+=damoteConfig.primaryComputationConfig.loadBal[i]
		*(1/(double)topo->nbLinks)*topo->nodeVec.cont[topo->nodeInd.cont[src]].info.sum[i]
		*topo->nodeVec.cont[topo->nodeInd.cont[src]].info.sum[i];
	}
	if (damoteConfig.primaryComputationConfig.load[i]!=0)
	{
	    score+=damoteConfig.primaryComputationConfig.load[i]*req->bw[i];
	    //score+=damoteConfig.primaryComputationConfig.load[i]*1;
	}
	if (damoteConfig.primaryComputationConfig.sqLoad[i]!=0)
	{
	    score+=damoteConfig.primaryComputationConfig.sqLoad[i]
		*(req->bw[i]*req->bw[i]+2*req->bw[i]*totBW[i]);
	}
	if (damoteConfig.primaryComputationConfig.relLoad[i]!=0)
	{
	    score+=damoteConfig.primaryComputationConfig.relLoad[i]
		*req->bw[i]/connect->linkState->cap[i];
	}
	if (damoteConfig.primaryComputationConfig.sqRelLoad[i]!=0)
	{
	    score+=damoteConfig.primaryComputationConfig.sqRelLoad[i]
		*(req->bw[i]*req->bw[i]+2*req->bw[i]*totBW[i])
		/(connect->linkState->cap[i]*connect->linkState->cap[i]);
	}
	if (damoteConfig.primaryComputationConfig.delay[i]!=0)
	{
	    score+=damoteConfig.primaryComputationConfig.delay[i]
		*((1/(connect->linkState->cap[i]-totBW[i]-req->bw[i]))
		  -(1/(connect->linkState->cap[i]-totBW[i])));
	}
    }
    
    for (i=0;i<NB_OA;i++)
    {    
	if (damoteConfig.primaryComputationConfig.rerouting[i]!=0)
	{
	    rerouteScore+=damoteConfig.primaryComputationConfig.rerouting[i]*
		makeRerouteScore(req,connect->info.gain,connect->linkState,i);
	}
    }

    score+=rerouteScore*(score>0?1:0)*score;
#if defined EBK
    fprintf(stderr,"Valeur de la fonction de score %f pour noeud source %ld noeud dest %ld\n", score, src, dst);
#endif
    return score;
}


int updateNodeInfoOnElect(BKTopology *topo,LSPRequest *req,long src,long dst,BKConnect *connect)
{
    long i;

    if (topo == NULL || connect == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    for (i=0;i<NB_OA;i++)
    {
	if (damoteConfig.primaryComputationConfig.loadBal[i]!=0)
	{
	    topo->nodeVec.cont[topo->nodeInd.cont[dst]].info.newSum[i]=
		topo->nodeVec.cont[topo->nodeInd.cont[src]].info.sum[i]+(req->bw[i]/connect->linkState->cap[i]);
	}
    }

#if defined EBK
    fprintf(stderr,"Sum value (new): %f\n", topo->nodeVec.cont[topo->nodeInd.cont[dst]].info.newSum[i]);
#endif 
    return 0;
}


int activateNodeInfo(BKTopology *topo,long nd)
{
    long i;

    if (topo == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    for (i=0;i<NB_OA;i++)
    {
	if (damoteConfig.primaryComputationConfig.loadBal[i]!=0)
	{
	    topo->nodeVec.cont[topo->nodeInd.cont[nd]].info.sum[i]=
		topo->nodeVec.cont[topo->nodeInd.cont[nd]].info.newSum[i];
	}
    }
#if defined EBK
    fprintf(stderr,"ActivateNodeInfo: sum value %f\n", topo->nodeVec.cont[topo->nodeInd.cont[nd]].info.sum[i]);
#endif
 
    return 0;
}


int noLoop(BKTopology *topo,long src,long dst)
{
    BKNode* tmpNode;
    

    tmpNode=&topo->nodeVec.cont[topo->nodeInd.cont[src]];
    while (tmpNode->neighbInd!=-1 && tmpNode->nodeId!=dst)
    {
	tmpNode=&topo->nodeVec.cont[topo->nodeInd.cont[tmpNode->inNeighb.cont[tmpNode->info.newNeighbInd].neighbId]];	
    }

    if (tmpNode->nodeId==dst)
	return 0;

    return 1;
}
