#include "database/database_api.h"
#include "database/database_util.h"

#include <stdio.h>
#include <string.h>

#if defined LINUX && defined TIMING
#include <sys/time.h>
#endif

/*******************************************
 * Communication structures and accessors: *
 *******************************************/


// Label switched path representation
//------------------------------------

DBLabelSwitchedPath* DBlspNew() 
{
    DBLabelSwitchedPath* lsp;

    if ((lsp=calloc(1,sizeof(DBLabelSwitchedPath)))==NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;	
    }
    
    if (longListInit(&(lsp->forbidLinks),-1)<0)
    {
	free(lsp);
	addError(CRITICAL,"Unable to create label switched path in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;
    }

    if (longListInit(&(lsp->path),-1)<0)
    {
	longListEnd(&(lsp->forbidLinks));
	free(lsp);
	addError(CRITICAL,"Unable to create label switched path in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;
    }   

    if (longListInit(&(lsp->primPath),-1)<0)
    {
	longListEnd(&(lsp->path));
	longListEnd(&(lsp->forbidLinks));
	free(lsp);
	addError(CRITICAL,"Unable to create label switched path in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;
    }   

    if (longListInit(&(lsp->backLSPIDs),-1)<0)
    {
	longListEnd(&(lsp->primPath));
	longListEnd(&(lsp->path));
	longListEnd(&(lsp->forbidLinks));
	free(lsp);
	addError(CRITICAL,"Unable to create label switched path in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;
    }     

    lsp->noContentionId=-1; //very important

    return lsp;
}

int DBlspInit(DBLabelSwitchedPath* lsp)
{
    if (lsp == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (longListInit(&(lsp->forbidLinks),-1)<0)
    {
	addError(CRITICAL,"Unable to create label switched path in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (longListInit(&(lsp->path),-1)<0)
    {
	longListEnd(&(lsp->forbidLinks));
	addError(CRITICAL,"Unable to create label switched path in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }   

    if (longListInit(&(lsp->primPath),-1)<0)
    {
	longListEnd(&(lsp->path));
	longListEnd(&(lsp->forbidLinks));
	addError(CRITICAL,"Unable to create label switched path in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }   

    if (longListInit(&(lsp->backLSPIDs),-1)<0)
    {
	longListEnd(&(lsp->primPath));
	longListEnd(&(lsp->path));
	longListEnd(&(lsp->forbidLinks));
	addError(CRITICAL,"Unable to create label switched path in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    memset(lsp->bw, 0, NB_OA * sizeof(double));
    lsp->noContentionId=-1;   //very important

    return 0;
}

int DBlspDestroy(DBLabelSwitchedPath* lsp)
{
    if (lsp == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
 
    longListEnd(&(lsp->backLSPIDs));
    longListEnd(&(lsp->primPath));
    longListEnd(&(lsp->path));
    longListEnd(&(lsp->forbidLinks));
    free(lsp);
    
    return 0;
}

int DBlspEnd(DBLabelSwitchedPath* lsp)
{
    if (lsp == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    longListEnd(&(lsp->backLSPIDs));
    longListEnd(&(lsp->primPath));
    longListEnd(&(lsp->path));
    longListEnd(&(lsp->forbidLinks));
    
    return 0;
}

int DBlspCopy(DBLabelSwitchedPath* dst, DBLabelSwitchedPath* src)
{
    int ret=0;

    if (dst == NULL || src==NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    dst->id=src->id;
    dst->precedence=src->precedence;
    memcpy(dst->bw,src->bw, NB_OA * sizeof(double));
    dst->noContentionId = src->noContentionId;
    ANDERROR(ret,longListCopy(&(dst->forbidLinks),&(src->forbidLinks)));
    ANDERROR(ret,longListCopy(&(dst->path),&(src->path)));
    dst->type=src->type;
    dst->primID=src->primID;
    ANDERROR(ret,longListCopy(&(dst->primPath),&(src->primPath)));
    ANDERROR(ret,longListCopy(&(dst->backLSPIDs),&(src->backLSPIDs)));
    
    if (ret<0)
    {
	addError(CRITICAL,"Label switched path copy uncomplete in %s at line %d",
		 __FILE__,__LINE__);
    }
    
    return ret;
}



// Ordered List of LSPs
//----------------------

DBLSPList* DBlspListNew(long size)
{
    DBLSPList *list=NULL;  
    void* ptr=NULL;
 
    if ((list = calloc(1,sizeof(DBLSPList))) == NULL) 
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;
    }

    if (size == -1)
	size = LSPLIST_INITSIZE;

    if ((ptr = calloc(size,sizeof(DBLabelSwitchedPath*))) == NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	free(list);
	return NULL;
    }
    
    list->size = size;
    list->top = 0;
    list->cont = ptr;

    return list;
}

int DBlspListInit(DBLSPList *list,long size)
{
    void* ptr=NULL;
    
    if (list == NULL)
    {
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (size == -1)
	size = LSPLIST_INITSIZE;

    if ((ptr = calloc(size,sizeof(DBLabelSwitchedPath*))) == NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    list->size = size;
    list->top = 0;
    list->cont = ptr;

    return 0;   
}

int DBlspListDestroy(DBLSPList *list)
{
    if (list == NULL || list->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    free(list->cont);
    free(list);

    return 0;
}

int DBlspListEnd(DBLSPList *list)
{
    if (list == NULL || list->cont == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    free(list->cont);
    list->cont = NULL;
    list->size = 0;
    list->top = 0;

    return 0;    
}

int DBlspListInsert(DBLSPList *list,DBLabelSwitchedPath *lsp)
{
    int a,b;
    void *ptr=NULL;

    if (list == NULL || list->cont == NULL || lsp == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    // check the capacity of the list
    if (list->top >= list->size)
    {
	if ((ptr = realloc(list->cont, list->size 
			   * 2 * sizeof(DBLabelSwitchedPath*))) == NULL) 
	{
	    addError(CRITICAL,"Critical lack of memory in %s at line %d",
		     __FILE__,__LINE__);
	    return -1;
	}
	else
	{
	    list->cont=ptr;
	    list->size*=2;
	}
    }

    // find the position in the list (to keep it sorted)
    a = 0;
    b = list->top-1;

    // empty list or after the last elem
    if (list->top == 0 || DBlspCompare(list->cont[b], lsp) >= 0) 
    {
	list->cont[list->top++] = lsp;
	return (list->top-1);
    }

    // before the first elem
    if (DBlspCompare(lsp, list->cont[a]) >= 0) 
    {
	memmove(list->cont+1, list->cont, (list->top)*sizeof(void*));
	list->cont[0] = lsp;
	list->top++;
	return 0;
    }

    // now the insert position is inside ]a,b[
    while (b - a > 1)
    {
	int mid = (a + b)/2;
	int ret = DBlspCompare(lsp, list->cont[mid]);

	if (ret == 1) 
	    b = mid;
	else if (ret == -1) 
	    a = mid;
	else // if (ret == 0)
	{
	    a = mid;
	    b = mid;
	}
    }

    // now insert before b
    memmove(list->cont+b+1, list->cont+b, (list->top - b)*sizeof(void*));
    list->cont[b] = lsp;
    list->top++;

    return b;
}

int DBlspCompare(const DBLabelSwitchedPath *LSPa,const DBLabelSwitchedPath *LSPb)
{
    if (LSPa->precedence > LSPb->precedence)
	return 1;
    else if (LSPa->precedence < LSPb->precedence)
	return -1;
    else if (LSPa->bw[0] > LSPb->bw[0])
	return 1;
    else if (LSPa->bw[0] < LSPb->bw[0])
	return -1;
    else 
    {
	if (LSPa->id < LSPb->id)
	    return 1;
	else if (LSPa->id > LSPb->id)
	    return -1;
    }

    return 0;
}

int DBlspListRemove(DBLSPList *list,DBLabelSwitchedPath *lsp)
{
    int a,b,index;

    if (list == NULL || list->cont == NULL || lsp == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    // find the position in the list
    a = 0;
    b = list->top-1;

    // empty list 
    if (list->top == 0) 
    {
	addError(WARNING,"Removing inexistent LSP in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    while (b - a > 1)
    {
	int mid = (a + b)/2;
	int ret = DBlspCompare(lsp, list->cont[mid]);
	
	if (ret == 1) 
	    b = mid;
	else if (ret == -1) 
	    a = mid;
	else // if (ret == 0)
	{
	    a = mid;
	    b = mid;
	}
    }

    if (DBlspCompare(lsp, list->cont[a]) == 0) 
    {
	index = a;
    }
    else if (DBlspCompare(lsp, list->cont[b]) == 0)
    {
	index = b;
    }
    else // not found
    {
	addError(WARNING,"Removing inexistent LSP in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    // now delete index
    memmove(list->cont + index, list->cont + index + 1, (list->top - index -1)*sizeof(void*));
    list->top--;

    return 0;
}



// Link state representation
//---------------------------

DBLinkState* DBlinkStateNew()
{
    DBLinkState* ls;
    int i,j,k,l;

    if ((ls=calloc(1,sizeof(DBLinkState)))==NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;	
    }
	
    for (i=0;i<NB_OA;i++)
	for (j=0;j<NB_PREEMPTION;j++)
	{	
	    if (dblVecInit(&(ls->bbw[i][j]),-1)<0)
	    {
		for (k=i;k>=0;k--)
		    for (l=j-1;l>=0;l--)
		    { 
			dblVecEnd(&(ls->bbw[k][l]));
			dblVecEnd(&(ls->remoteBbw[k][l]));
			dblVecEnd(&(ls->fbw[k][l]));
			dblVecEnd(&(ls->remoteFbw[k][l]));
		    }
		free(ls);
		addError(CRITICAL,"Unable to create link state in %s at line %d",
		 __FILE__,__LINE__);
		return NULL;
	    }
	    else if (dblVecInit(&(ls->remoteBbw[i][j]),-1)<0)
	    {
		dblVecEnd(&(ls->bbw[i][j]));
		for (k=i;k>=0;k--)
		    for (l=j-1;l>=0;l--)
		    { 
			dblVecEnd(&(ls->bbw[k][l]));
			dblVecEnd(&(ls->remoteBbw[k][l]));
			dblVecEnd(&(ls->fbw[k][l]));
			dblVecEnd(&(ls->remoteFbw[k][l]));
		    }
		free(ls);
		addError(CRITICAL,"Unable to create link state in %s at line %d",
			 __FILE__,__LINE__);
		return NULL;
	    }
	    else if (dblVecInit(&(ls->fbw[i][j]),-1)<0)
	    {
		dblVecEnd(&(ls->bbw[i][j]));
		dblVecEnd(&(ls->remoteBbw[i][j]));
		for (k=i;k>=0;k--)
		    for (l=j-1;l>=0;l--)
		    { 
			dblVecEnd(&(ls->bbw[k][l]));
			dblVecEnd(&(ls->remoteBbw[k][l]));
			dblVecEnd(&(ls->fbw[k][l]));
			dblVecEnd(&(ls->remoteFbw[k][l]));
		    }
		free(ls);
		addError(CRITICAL,"Unable to create link state in %s at line %d",
			 __FILE__,__LINE__);
		return NULL;
	    }
	    else if (dblVecInit(&(ls->remoteFbw[i][j]),-1)<0)
	    {
		dblVecEnd(&(ls->bbw[i][j]));
		dblVecEnd(&(ls->remoteBbw[i][j]));
		dblVecEnd(&(ls->fbw[i][j]));
		for (k=i;k>=0;k--)
		    for (l=j-1;l>=0;l--)
		    { 
			dblVecEnd(&(ls->bbw[k][l]));
			dblVecEnd(&(ls->remoteBbw[k][l]));
			dblVecEnd(&(ls->fbw[k][l]));
			dblVecEnd(&(ls->remoteFbw[k][l]));
		    }
		free(ls);
		addError(CRITICAL,"Unable to create link state in %s at line %d",
			 __FILE__,__LINE__);
		return NULL;
	    } 
	}

    return ls;
}

int DBlinkStateInit(DBLinkState* ls)
{
    int i,j,k,l;

    if (ls == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    memset(ls, 0, sizeof(DBLinkState));

    for (i=0;i<NB_OA;i++)
	for (j=0;j<NB_PREEMPTION;j++)
	{	
	    if (dblVecInit(&(ls->bbw[i][j]),-1)<0)
	    {
		for (k=i;k>=0;k++)
		    for (l=j-1;l>=0;l++)
		    { 
			dblVecEnd(&(ls->bbw[k][l]));
			dblVecEnd(&(ls->remoteBbw[k][l]));
			dblVecEnd(&(ls->fbw[k][l]));
			dblVecEnd(&(ls->remoteFbw[k][l]));
		    }
		addError(CRITICAL,"Unable to create link state in %s at line %d",
			 __FILE__,__LINE__);
		return -1;
	    }
	    else if (dblVecInit(&(ls->remoteBbw[i][j]),-1)<0)
	    {
		dblVecEnd(&(ls->bbw[i][j]));
		for (k=i;k>=0;k++)
		    for (l=j-1;l>=0;l++)
		    { 
			dblVecEnd(&(ls->bbw[k][l]));
			dblVecEnd(&(ls->remoteBbw[k][l]));
			dblVecEnd(&(ls->fbw[k][l]));
			dblVecEnd(&(ls->remoteFbw[k][l]));
		    }
		addError(CRITICAL,"Unable to create link state in %s at line %d",
			 __FILE__,__LINE__);
		return -1;
	    }
	    else if (dblVecInit(&(ls->fbw[i][j]),-1)<0)
	    {
		dblVecEnd(&(ls->bbw[i][j]));
		dblVecEnd(&(ls->remoteBbw[i][j]));
		for (k=i;k>=0;k++)
		    for (l=j-1;l>=0;l++)
		    { 
			dblVecEnd(&(ls->bbw[k][l]));
			dblVecEnd(&(ls->remoteBbw[k][l]));
			dblVecEnd(&(ls->fbw[k][l]));
			dblVecEnd(&(ls->remoteFbw[k][l]));
		    }
		addError(CRITICAL,"Unable to create link state in %s at line %d",
			 __FILE__,__LINE__);
		return -1;
	    }
	    else if (dblVecInit(&(ls->remoteFbw[i][j]),-1)<0)
	    {
		dblVecEnd(&(ls->bbw[i][j]));
		dblVecEnd(&(ls->remoteBbw[i][j]));
		dblVecEnd(&(ls->fbw[i][j]));
		for (k=i;k>=0;k++)
		    for (l=j-1;l>=0;l++)
		    { 
			dblVecEnd(&(ls->bbw[k][l]));
			dblVecEnd(&(ls->remoteBbw[k][l]));
			dblVecEnd(&(ls->fbw[k][l]));
			dblVecEnd(&(ls->remoteFbw[k][l]));
		    }
		addError(CRITICAL,"Unable to create link state in %s at line %d",
			 __FILE__,__LINE__);
		return -1;
	    }
	}
    
    return 0;
}

int DBlinkStateDestroy(DBLinkState* ls)
{
    int i,j;

    if (ls == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    for (i=0;i<NB_OA;i++)
	for (j=0;j<NB_PREEMPTION;j++)
	{
	    dblVecEnd(&(ls->bbw[i][j]));
	    dblVecEnd(&(ls->remoteBbw[i][j]));
	    dblVecEnd(&(ls->fbw[i][j]));
	    dblVecEnd(&(ls->remoteFbw[i][j]));
	}
    free(ls);

    return 0;
}

int DBlinkStateEnd(DBLinkState* ls)
{
    int i,j;

    if (ls == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    for (i=0;i<NB_OA;i++)
	for (j=0;j<NB_PREEMPTION;j++)
	{
	    dblVecEnd(&(ls->bbw[i][j]));
	    dblVecEnd(&(ls->remoteBbw[i][j]));
	    dblVecEnd(&(ls->fbw[i][j]));
	    dblVecEnd(&(ls->remoteFbw[i][j]));
	}
    
    return 0;
}

int DBlinkStateCopy(DBLinkState* dst, DBLinkState* src)
{
    int i,j,ret=0;
    
    if (dst == NULL || src == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    dst->color=src->color;
    memcpy(&(dst->cap),&(src->cap),NB_OA * sizeof(double));
    memcpy(&(dst->rbw),&(src->rbw),NB_OA * NB_PREEMPTION * sizeof(double));
    memcpy(&(dst->pbw),&(src->pbw),NB_OA * NB_PREEMPTION * sizeof(double));

    for (i=0;(i<NB_OA && ret>=0);i++)
	for (j=0;(j<NB_PREEMPTION && ret>=0);j++)
	{
	    ANDERROR(ret,dblVecCopy(&(dst->bbw[i][j]),&(src->bbw[i][j])));
	    ANDERROR(ret,dblVecCopy(&(dst->remoteBbw[i][j]),&(src->remoteBbw[i][j])));
	    ANDERROR(ret,dblVecCopy(&(dst->fbw[i][j]),&(src->fbw[i][j])));
	    ANDERROR(ret,dblVecCopy(&(dst->remoteFbw[i][j]),&(src->remoteFbw[i][j])));
	}
    
    if (ret<0)
    {
	addError(CRITICAL,"Link state copy uncomplete in %s at line %d",
		 __FILE__,__LINE__);
    }
    
    return ret;
}

int computeRBW(DataBase* dataBase,double rbw[NB_OA][NB_PREEMPTION], double pbw[NB_OA][NB_PREEMPTION], 
	       DoubleVec bbw[NB_OA][NB_PREEMPTION], DoubleVec fbw[NB_OA][NB_PREEMPTION])
{
#if defined LINUX && defined TIME1
    struct timezone tz;
    struct timeval  t1,t2;
#endif
    int nbLink = 0, seenLinks;
    int nbNode = 0, seenNodes;
    int i,oa,p;
    DoubleVec* gbw;
    double totBbw = 0;
    double totFbw = 0;
    double m, oldM;

    
    

    if ((rbw==NULL) || (pbw==NULL) || (bbw==NULL) || (fbw==NULL))
    {
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

#if defined LINUX && defined TIME1
    gettimeofday(&t1, &tz);
#endif

    nbLink = dataBase->linkSrcVec.top;
    nbNode = dataBase->nodeVec.top;

    if ((gbw = calloc(nbLink + nbNode, sizeof(DoubleVec))) == NULL)
    {
	addError(CRITICAL,"Cannot allocate GBW in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    for (i=0; i<nbLink + nbNode; ++i)
	dblVecInit(&(gbw[i]), NB_PREEMPTION); 

    for (oa=0; oa<NB_OA; ++oa)
    {
	seenLinks = 0;

	// phase 1a (links)
	for (i=0; seenLinks<dataBase->nbLinks; ++i)
	{
	    if (dataBase->linkSrcVec.cont[i] == 0)
		continue;
	    else
		seenLinks++;

	    totBbw = 0;
	    totFbw = 0;

	    for(p=0; p<NB_PREEMPTION; ++p)
	    {
		if (i < bbw[oa][p].size)
		    totBbw += bbw[oa][p].cont[i];
		if (i < fbw[oa][p].size)
		    totFbw += fbw[oa][p].cont[i];

		gbw[i].cont[p] = max(0, totBbw - totFbw);  
	    }
	}

	seenNodes = 0;

	// phase 1b (nodes or any set of links)
	for (i=0; seenNodes<dataBase->nbNodes; ++i)
	{
	    if (dataBase->nodeVec.cont[i] == NULL)
		continue;
	    else
		seenNodes++;

	    totBbw = 0;
	    totFbw = 0;

	    for(p=0; p<NB_PREEMPTION; ++p)
	    {
		LongList* lst;
		if ((lst = DBgetNodeInNeighb(dataBase, i)) != NULL)
		{
		    int l;
		    for (l=0; l<lst->top; ++l)
		    {
			int lnkID = DBgetLinkID(dataBase, lst->cont[l], i);
			if (lnkID < bbw[oa][p].size)
			    totBbw += bbw[oa][p].cont[lnkID];
			if (lnkID < fbw[oa][p].size)
			    totFbw += fbw[oa][p].cont[lnkID];
		    }
		}

		gbw[i + nbLink].cont[p] = max(0, totBbw - totFbw);  
	    }
	}

	// phase 2
	oldM = 0;

	for (p=0; p<NB_PREEMPTION; ++p)
	{
	    m = 0;
	    for (i=0; i<nbLink + nbNode; ++i)
	    {
		if (gbw[i].cont[p] > m)
		{
		    m = gbw[i].cont[p];
		}
	    }

	    rbw[oa][p] = pbw[oa][p] + m - oldM;
	    oldM = m;
	}
    }

    for (i=0; i<nbLink + nbNode; ++i)
	dblVecEnd(&(gbw[i]));

    free(gbw);

#if defined LINUX && defined TIME1
    gettimeofday(&t2, &tz);
    fprintf(stderr, "Time to compute rbw : %f ms\n", (t2.tv_sec - t1.tv_sec) * 1000 + 
	    (t2.tv_usec - t1.tv_usec) / 1000.0);
#endif

   

    return 0;

}

typedef enum operation_ { SETUP, REMOVE} operation;

int updateLS(DataBase*, long, long, DBLinkState*, DBLabelSwitchedPath*, operation);

int evalLS(DataBase* dataBase, long src, long dst, DBLinkState* newLS, DBLinkState* oldLS, LSPRequest* req, operation op)
{
    DBLabelSwitchedPath* lsp, *primLSP;
    int ret;

    // check the arguments
    if ((dataBase==NULL) || (newLS==NULL) || (oldLS==NULL) || (req==NULL))
    {
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    // duplicate the LS
    if (newLS != oldLS && DBlinkStateCopy(newLS, oldLS) < 0)
    {
	addError(CRITICAL,"Impossible to duplicate the linkState in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    // now build a false LSP satisfying the request ....
    lsp = DBlspNew();
    lsp->id = req->id;
    lsp->precedence = req->precedence;
    memcpy(lsp->bw, req->bw, NB_OA * sizeof(double));
    longListCopy(&(lsp->forbidLinks), &(req->forbidLinks));

    if (req->rerouteInfo.id >= 0)
    {
	lsp->noContentionId = req->rerouteInfo.id;
    }

    switch(req->type)
    {
	case PRIM:
	    lsp->type = PRIM;
	    lsp->primID = -1;
	    break;

	case GLOBAL_BACK:
	case LOCAL_BACK:
	    lsp->type = req->type;
	    lsp->primID = req->primID;
	    
	    // look up the primary path ....
	    if ((primLSP = DBgetLSP(dataBase, lsp->primID)) == NULL)
	    {
		addError(CRITICAL,"Impossible to determine the primary path in %s at line %d",
		 __FILE__,__LINE__);
		DBlspDestroy(lsp);
		return -1;
	    }
	    
	    longListCopy(&(lsp->primPath), &(primLSP->path));

	    break;

	default:
	    addError(CRITICAL,"Unknown request type (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	    DBlspDestroy(lsp);
	    return -1;
    }

    if (req->path.top < 2)
    {
	addError(CRITICAL,"Wrong path in request in %s at line %d",
		 __FILE__,__LINE__);
	DBlspDestroy(lsp);
	return -1;	
    }

    if (longListCopy(&(lsp->path), &(req->path)) < 0)
    {
	addError(CRITICAL,"Impossible to duplicate path in %s at line %d",
		 __FILE__,__LINE__);
	DBlspDestroy(lsp);
	return -1;
    }

    ret = updateLS(dataBase, src, dst, newLS, lsp, op);

    // clean up ....
    DBlspDestroy(lsp);

    return ret;
}

int updateLS(DataBase* dataBase, long src, long dst, DBLinkState* ls, DBLabelSwitchedPath* lsp, operation op)
{
    bool path = FALSE;
    int myPosPath=-1, myPosPrimPath=-1;
    bool primPath = FALSE;
    int i,j,oa;
    int mult;
    int plink;
    double newBW[NB_OA];
    DBLabelSwitchedPath* oldLSP=NULL;
    bool rerouting = FALSE;
    LongVec idList;


    // check the arguments
    if ((dataBase==NULL) || (ls==NULL) || (lsp==NULL))
    {
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    // am i on the path ?
    for (i=0; i<lsp->path.top; ++i)
    {
	if (lsp->path.cont[i] == src)
	    break;
    }

    if (i < (lsp->path.top - 1) && lsp->path.cont[i+1] == dst)
    {
	path = TRUE;
	myPosPath = i;
    }
    
    // if i am on the path, am i the a rerouted LSP sharing the link with a preempted one ?
    if (path == TRUE && lsp->noContentionId >= 0)
    {
	if ((oldLSP = DBgetLSP(dataBase, lsp->noContentionId)) == NULL) 
	{
	    addError(WARNING,"Cannot find the old LSP in %s at line %d",
		 __FILE__,__LINE__);
	}
	else
	{
	    for (i=0; i<oldLSP->path.top; ++i)
	    {
		if (oldLSP->path.cont[i] == src)
		    break;
	    }
	    
	    if (i < (oldLSP->path.top - 1) && oldLSP->path.cont[i+1] == dst)
	    {
		rerouting = TRUE;
	    }
	}
    }

/*
    // if rerouting -> check if there is a change in the bandwidth reservation. if no change we can leave
    if (rerouting == TRUE) 
    {
	bool test=FALSE;

	
	for (i=0; i<NB_OA; i++) {
	    if ((newBW[i] = max(lsp->bw[i] - oldLSP->bw[i], 0)) != 0)
		test = TRUE;
	}

	if (test == FALSE) 
	{
	    return 0;
	}
    }
*/

    // if it is a backup am i on the prim path ?
    if (lsp->type == LOCAL_BACK || lsp->type == GLOBAL_BACK)
    {
	for (i=0; i<lsp->primPath.top; ++i)
	{
	    if (lsp->primPath.cont[i] == src)
		break;
	}

	if (i < (lsp->primPath.top - 1) && lsp->primPath.cont[i+1] == dst)
	{
	    primPath = TRUE;
	    myPosPrimPath = i;
	}
    }

    if (!path && !primPath) // not concerned by this update ...
	return 0;

    if (op == SETUP)
    {
	mult = 1;
    }
    else if (op == REMOVE)
    {
	mult = -1;
    }
    else 
    {
	addError(CRITICAL,"Bad argument (unknown operation) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if (lsp->type == LOCAL_BACK)
    {
	// the path is a local backup
	// --------------------------
	int start,end;

	// which link are we protecting ... and the start
	for (i=0; i<lsp->primPath.top; ++i)
	{
	    if (lsp->primPath.cont[i] == lsp->path.cont[0])
		break;
	}

	if (i < lsp->primPath.top - 1)
	{
	    plink = DBgetLinkID(dataBase, lsp->path.cont[0], lsp->primPath.cont[i+1]);
	    start = i;
	}
	else 
	{
	  /*   fprintf(stderr,"\nChemin du primaire %ld: ", lsp->primID);
	    int j;
	    for (j=0;j<lsp->primPath.top;j++){
		fprintf(stderr,"%ld ", lsp->primPath.cont[j]);
	    }
	    fprintf(stderr,"\nChemin du backup %ld: ", lsp->id);
	    for (j=0;j<lsp->path.top;j++){
		fprintf(stderr,"%ld ", lsp->path.cont[j]);
		}*/
	    
	    addError(CRITICAL,"Cannot determine the link to protect in %s at line %d",
		     __FILE__,__LINE__);
	    return -1;
	}

	if (path == TRUE)
	{
	    
	   
// update bbw
	    for (oa=0; oa<NB_OA; ++oa)
	    {
		if (ls->bbw[oa][lsp->precedence].size <= plink)
		    dblVecResize(&(ls->bbw[oa][lsp->precedence]), plink+1);

		if (rerouting == FALSE)
		    ls->bbw[oa][lsp->precedence].cont[plink] += (mult * lsp->bw[oa]);
		else
		    ls->bbw[oa][lsp->precedence].cont[plink] += (mult * newBW[oa]);
	    }
	}

	if (primPath == TRUE)
	{
	    // find the end
	    for (i=start; i<lsp->primPath.top; ++i)
	    {
		if (lsp->primPath.cont[i] == lsp->path.cont[lsp->path.top - 1])
		    break;
	    }
	    
	    if (i < lsp->primPath.top)
	    {
		end = i;
	    }
	    else 
	    {
		// finding an end is not required during path evaluation
		end = -1;
		/*
		addError(CRITICAL,"Cannot determine the merging point in %s at line %d",
			 __FILE__,__LINE__);
		return -1;
		*/
	    }

	    // maybe fbw must be recomputed
	    // if i'm before path.cont[0] or after path.cont[last]
	    if (myPosPrimPath < start || myPosPrimPath >= end)
	    {
		for (oa=0; oa<NB_OA; ++oa)
		{
		    if (ls->fbw[oa][lsp->precedence].size <= plink)
			dblVecResize(&(ls->fbw[oa][lsp->precedence]), plink+1);

		    if (rerouting == FALSE) 
			ls->fbw[oa][lsp->precedence].cont[plink] -= (mult * lsp->bw[oa]);
		    else
			ls->fbw[oa][lsp->precedence].cont[plink] -= (mult * newBW[oa]);
		}
	    }
	}
    }
    else if (lsp->type == GLOBAL_BACK)
    {
	// the path is a end-to-end backup
	// -------------------------------
	int start,end;

	// we are protecting all nodes between path.cont[0] and path.cont[end]
	// find the start
	for (i=0; i<lsp->primPath.top; ++i)
	{
	    if (lsp->primPath.cont[i] == lsp->path.cont[0])
		break;
	}

	if (i < lsp->primPath.top)
	{
	    start = i;
	}
	else 
	{
	    addError(CRITICAL,"Cannot determine the link to protect in %s at line %d",
		 __FILE__,__LINE__);
	    return -1;
	}

	// find the end
	end = lsp->primPath.top - 1;

	if (path == TRUE)
	{
	    // update bbw
	    for (i=start; i<end; ++i)
	    {
		if ((plink = DBgetLinkID(dataBase, lsp->primPath.cont[i], lsp->primPath.cont[i+1])) >= 0)
                    for (oa=0; oa<NB_OA; ++oa)
                    {
                        if (ls->bbw[oa][lsp->precedence].size <= plink)
                            dblVecResize(&(ls->bbw[oa][lsp->precedence]), plink+1);

			if (rerouting == FALSE)
			    ls->bbw[oa][lsp->precedence].cont[plink] += (mult * lsp->bw[oa]);
			else
			    ls->bbw[oa][lsp->precedence].cont[plink] += (mult * newBW[oa]);
                    }
                else
                {
                    addError(CRITICAL,"Cannot determine link ID in %s at line %d",
                             __FILE__,__LINE__);
                    return -1;
                }
	    }
	}

	if (primPath == TRUE)
	{
	    // maybe fbw must be recomputed
	    // if i'm before path.cont[0] or after path.cont[last]
	    
	    for (i=0; i<start; ++i)
	    {
                if ((plink = DBgetLinkID(dataBase, lsp->primPath.cont[i], lsp->primPath.cont[i+1])) >= 0)
                    for (oa=0; oa<NB_OA; ++oa)
                    {
                        if (ls->fbw[oa][lsp->precedence].size <= plink)
                            dblVecResize(&(ls->fbw[oa][lsp->precedence]), plink+1);

			if (rerouting == FALSE)
			    ls->fbw[oa][lsp->precedence].cont[plink] -= (mult * lsp->bw[oa]);
			else
			    ls->fbw[oa][lsp->precedence].cont[plink] -= (mult * newBW[oa]);
                    }
                else
                {
                    addError(CRITICAL,"Cannot determine link ID in %s at line %d",
                             __FILE__,__LINE__);
                    return -1;
                }
	    }

	    for (i=end; i<lsp->primPath.top-1; ++i)
	    {
		plink = DBgetLinkID(dataBase, lsp->primPath.cont[i], lsp->primPath.cont[i+1]);
		for (oa=0; oa<NB_OA; ++oa)
		{
		    if (ls->fbw[oa][lsp->precedence].size <= plink)
			dblVecResize(&(ls->fbw[oa][lsp->precedence]), plink+1);

		    if (rerouting == FALSE)
			ls->fbw[oa][lsp->precedence].cont[plink] -= (mult * lsp->bw[oa]);
		    else
			ls->fbw[oa][lsp->precedence].cont[plink] -= (mult * newBW[oa]);
		}
	    }
	}
    }
    else 
    {
	// path is a primary

	// update pbw
	for (i=0; i<NB_OA; ++i)
	{
	    if (rerouting == FALSE)
		ls->pbw[i][lsp->precedence] += (mult * lsp->bw[i]);
	    else
		ls->pbw[i][lsp->precedence] += (mult * newBW[i]);
	}

	// update fbw

	if (oldLSP != NULL) {

	    longVecInit(&idList, oldLSP->path.top);

	    for (j=0; j<oldLSP->path.top - 1; ++j) {
		int id;
		if ((id = DBgetLinkID(dataBase, oldLSP->path.cont[j], oldLSP->path.cont[j+1])) >= 0)
		{
		    longListPushBack(&idList, id);
		}
		else
		{
		    addError(CRITICAL,"Cannot determine link ID in %s at line %d",
			     __FILE__,__LINE__);
		    return -1;
		}
	    }
	}

	for (i=0; i<lsp->path.top - 1; ++i)
	{
            int id;

            if ((id = DBgetLinkID(dataBase, lsp->path.cont[i], lsp->path.cont[i+1])) >= 0)
                for (oa=0; oa<NB_OA; ++oa)
                {
		    bool isCommon;

                    if (ls->fbw[oa][lsp->precedence].size <= id)
                        dblVecResize(&(ls->fbw[oa][lsp->precedence]), id+1);

		    isCommon = FALSE;
		    
		    if (oldLSP != NULL) {
			for (j=0; j<idList.top; ++j) {
			    if (idList.cont[j] == id) {
				isCommon = TRUE;
				break;
			    }
			}
		    }

		    if (rerouting == FALSE || isCommon == FALSE)
			ls->fbw[oa][lsp->precedence].cont[id] += (mult * lsp->bw[oa]);
		    else
			ls->fbw[oa][lsp->precedence].cont[id] += (mult * newBW[oa]);
                }
            else
            {
                addError(CRITICAL,"Cannot determine link ID in %s at line %d",
                         __FILE__,__LINE__);
                return -1;
            }
	}
	
	if (oldLSP != NULL)
	    longVecEnd(&idList);
    }

    computeRBW(dataBase, ls->rbw, ls->pbw, ls->bbw, ls->fbw); 

    return 0;
}

int DBevalLSOnSetup(DataBase* dataBase, long src, long dst, DBLinkState* newLS, DBLinkState* oldLS, LSPRequest* req)
{
    return evalLS(dataBase, src, dst, newLS, oldLS, req, SETUP);
}


int DBevalLSOnRemove(DataBase* dataBase, long src, long dst, DBLinkState* newLS, DBLinkState* oldLS, LSPRequest* req)
{
    return evalLS(dataBase, src, dst, newLS, oldLS, req, REMOVE);
}

int DBupdateLSOnSetup(DataBase* dataBase, long src, long dst, DBLinkState* ls, DBLabelSwitchedPath* lsp)
{
    return updateLS(dataBase, src, dst, ls, lsp, SETUP);
}

int DBupdateLSOnRemove(DataBase* dataBase, long src, long dst, DBLinkState* ls, DBLabelSwitchedPath* lsp)
{
    return updateLS(dataBase, src, dst, ls, lsp, REMOVE);
}


/**********************
 *    database API:   *
 **********************/


DataBase *DBnew(long ID)
{
    DataBase *dataBase=NULL;

    if ((dataBase=calloc(1,sizeof(DataBase)))==NULL)
    {
	addError(CRITICAL,"Critical lack of memory in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;
    }

    dataBase->id=ID;

    if (DBnodeVecInit(&(dataBase->nodeVec),-1)<0)
    {	
	addError(CRITICAL,"Unable to initialize the general node container in %s at line %d",
		 __FILE__,__LINE__);
	free(dataBase);
	return NULL;
    }

    if (DBlspVecInit(&(dataBase->lspVec),-1)<0)
    {
	addError(CRITICAL,"Unable to initialize the general LSP container in %s at line %d",
		 __FILE__,__LINE__);
	DBnodeVecEnd(&(dataBase->nodeVec));
	free(dataBase);
	return NULL;
    }

    if (DBlinkTabInit(&(dataBase->linkTab),-1)<0)
    {
	addError(CRITICAL,"Unable to initialize the general link container in %s at line %d",
		 __FILE__,__LINE__);
	DBnodeVecEnd(&(dataBase->nodeVec));
	DBlspVecEnd(&(dataBase->lspVec));
	free(dataBase);
	return NULL;
    }

    if (longVecInit(&(dataBase->linkSrcVec),LINKTAB_INITSIZE)<0)
    {
	addError(CRITICAL,"Unable to initialize the link id-src translater in %s at line %d",
		 __FILE__,__LINE__);
	DBnodeVecEnd(&(dataBase->nodeVec));
	DBlspVecEnd(&(dataBase->lspVec));
	DBlinkTabEnd(&(dataBase->linkTab));
	free(dataBase);
	return NULL;
    }
	
    if (longVecInit(&(dataBase->linkDstVec),LINKTAB_INITSIZE)<0)
    {
	addError(CRITICAL,"Unable to initialize the link id-dst translater in %s at line %d",
		 __FILE__,__LINE__);
	DBnodeVecEnd(&(dataBase->nodeVec));
	DBlspVecEnd(&(dataBase->lspVec));
	DBlinkTabEnd(&(dataBase->linkTab));
	longVecEnd(&(dataBase->linkSrcVec));
	free(dataBase);
	return NULL;
    }    

    dataBase->nbNodes=0;
    dataBase->nbLinks=0;

    return dataBase;
}

int DBdestroy(DataBase *dataBase)
{    
    if (dataBase == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    DBnodeVecEnd(&(dataBase->nodeVec));
    DBlspVecEnd(&(dataBase->lspVec));
    DBlinkTabEnd(&(dataBase->linkTab));
    longVecEnd(&(dataBase->linkSrcVec));
    longVecEnd(&(dataBase->linkDstVec));

    free(dataBase);
    
    return 0;
}

long DBgetID(DataBase *dataBase)
{
    if (dataBase == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    return dataBase->id;
}

long DBgetLinkID(DataBase *dataBase,long src,long dst)
{
    DBLink *lnk=NULL;

    if (dataBase == NULL || src < 0 || dst < 0)
    { 
	addError(CRITICAL,"Bad argument (NULL or negative value) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if ((lnk=DBlinkTabGet(&(dataBase->linkTab),src,dst))==NULL)
    {
	return -1;
    }

    return lnk->id;
}

long DBgetLinkSrc(DataBase *dataBase,long id)
{
    long ret;

    if (dataBase == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    if (longVecGet(&(dataBase->linkSrcVec),id,&ret)<0)
    {
	addError(CRITICAL,"Inexistent link in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    return (ret-1);	
}

long DBgetLinkDst(DataBase *dataBase,long id)
{
    long ret;

    if (dataBase == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    if (longVecGet(&(dataBase->linkDstVec),id,&ret)<0)
    {
	addError(CRITICAL,"Inexistent link in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    return (ret-1);
}

long DBgetNbNodes(DataBase *dataBase)
{
    if (dataBase == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    return dataBase->nbNodes;
}

long DBgetMaxNodeID(DataBase *dataBase)
{
    if (dataBase == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    return dataBase->nodeVec.top-1;
}

int DBaddNode(DataBase *dataBase,long id)
{
    DBNode *node=NULL;

    if (dataBase == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if ((node=DBnodeNew()) == NULL) 
    {
	addError(CRITICAL,"Unable to create node in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    node->id=id;
    
    if (DBnodeVecSet(&(dataBase->nodeVec),node,id) < 0)
    {
	addError(CRITICAL,"Unable to insert a new node in the general node container in %s at line %d",
		 __FILE__,__LINE__);
	DBnodeDestroy(node);
	return -1;
    }

    dataBase->nbNodes++;

    return 0;
}

int DBremoveNode(DataBase *dataBase, long id)
{
    DBNode *node=NULL;
    int ret=0;
    
    if (dataBase == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if ((node=DBnodeVecGet(&(dataBase->nodeVec),id)) == NULL)
    {
	addError(CRITICAL,"Trying to remove an inexistent node in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    // remember that DBremoveLink will update the neighbour list 
    while(node->inNeighb.top > 0)
    {
	ANDERROR(ret,DBremoveLink(dataBase,node->inNeighb.cont[node->inNeighb.top-1],id));
    }
    
    // remember that DBremoveLink will update the neighbour list 
    while(node->outNeighb.top > 0)
    {
	ANDERROR(ret,DBremoveLink(dataBase,id,node->outNeighb.cont[node->outNeighb.top-1]));
    }    

    ANDERROR(ret,DBnodeVecRemove(&(dataBase->nodeVec),id));

    if (ret<0)
    {
	addError(CRITICAL,"Node removal uncomplete in %s at line %d",
		 __FILE__,__LINE__);
    }

    dataBase->nbLinks--;

    return ret;
}

long DBgetNbLinks(DataBase *dataBase)
{
    if (dataBase == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    return dataBase->nbLinks;
}

int DBaddLink(DataBase *dataBase,long id,long src,long dst, DBLinkState* initLinkState)
{
    DBLink* link=NULL;
    int ret=0;

    if (dataBase == NULL || initLinkState==NULL 
	|| id <0 || src<0 || dst<0) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    if (((id<dataBase->linkSrcVec.size) && (dataBase->linkSrcVec.cont[id]>0))
	||
	((id<dataBase->linkDstVec.size) && (dataBase->linkDstVec.cont[id]>0)))
    {
	addError(CRITICAL,"Trying to add a link with a reserved ID (ID=%ld) in %s at line %d",
		 id,__FILE__,__LINE__);
	return -1;
    }	
    
    if ((link = DBlinkNew()) == NULL)
    {
	addError(CRITICAL,"Unable to create link in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    link->id=id;

    if (DBlinkStateCopy(&(link->state), initLinkState))
    {
	addError(CRITICAL,"Unable to create link in %s at line %d",
		 __FILE__,__LINE__);
	DBlinkDestroy(link);
	return -1;
    }

    if ((DBnodeVecGet(&(dataBase->nodeVec),src) == NULL) ||
	(DBnodeVecGet(&(dataBase->nodeVec),dst) == NULL))
    {
	addError(CRITICAL,"Source or destination node doesn't exist in %s at line %d",
		 __FILE__,__LINE__);
	DBlinkDestroy(link);
	return -1;
    }

    if (DBlinkTabSet(&(dataBase->linkTab),link,src,dst)<0)
    {
	addError(CRITICAL,"Unable to insert a new node in the general node container in %s at line %d",
		 __FILE__,__LINE__);
	DBlinkDestroy(link);
	return -1;
    }

    ANDERROR(ret,longListPushBack(&(dataBase->nodeVec.cont[src]->outNeighb),dst));
    ANDERROR(ret,longListPushBack(&(dataBase->nodeVec.cont[dst]->inNeighb),src));

    ANDERROR(ret,longListSort(&(dataBase->nodeVec.cont[src]->outNeighb)));
    ANDERROR(ret,longListSort(&(dataBase->nodeVec.cont[dst]->inNeighb)));

    ANDERROR(ret,longVecSet(&(dataBase->linkSrcVec),id,src+1));
    ANDERROR(ret,longVecSet(&(dataBase->linkDstVec),id,dst+1));

    // Maximum non-null element
    dataBase->linkSrcVec.top = max(dataBase->linkSrcVec.top, id+1);
    dataBase->linkDstVec.top = dataBase->linkSrcVec.top;

    if (ret<0)
    {
	addError(CRITICAL,"Link addition uncomplete in %s at line %d",
		 __FILE__,__LINE__);
    }

    dataBase->nbLinks++;

    return ret;
}

int DBremoveLink(DataBase *dataBase,long src,long dst)
{
    

    int id,ret=0;
    int i=0;
    int srcindex = 0;
    int dstindex = 0;
    LongList* outNeighbours = NULL;
    LongList* inNeighbours = NULL;

    if (dataBase == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if ((DBnodeVecGet(&(dataBase->nodeVec),src)==NULL) ||
	(DBnodeVecGet(&(dataBase->nodeVec),dst)==NULL) ||
	(DBlinkTabGet(&(dataBase->linkTab),src,dst)==NULL))
    {
	addError(CRITICAL,"Link doesn't exist or database unconsistancy in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    id=DBgetLinkID(dataBase,src,dst);

    /** Code added by Olivier Delcourt to retrieve the correct index !
	Should add error detection...
     **/
    outNeighbours = &(dataBase->nodeVec.cont[src]->outNeighb);
    inNeighbours = &(dataBase->nodeVec.cont[dst]->inNeighb);
    
    for (i=0; i<outNeighbours->top; i++){
	if (outNeighbours->cont[i]==dst){
	    dstindex = i;
	    break;
	}
    }

    for (i=0; i<inNeighbours->top; i++){
	if (inNeighbours->cont[i]==src){
	    srcindex = i;
	    break;
	}
    }

    ANDERROR(ret,longListRemove(&(dataBase->nodeVec.cont[src]->outNeighb),dstindex));
  
    ANDERROR(ret,longListRemove(&(dataBase->nodeVec.cont[dst]->inNeighb),srcindex));
  
    ANDERROR(ret,DBlinkTabRemove(&(dataBase->linkTab),src,dst));
   
    ANDERROR(ret,longVecSet(&(dataBase->linkSrcVec),id,0));
     
    ANDERROR(ret,longVecSet(&(dataBase->linkDstVec),id,0));
   
    while (dataBase->linkSrcVec.cont[dataBase->linkSrcVec.top-1] == 0)
	dataBase->linkSrcVec.top--;

    if (ret<0)
    {
	addError(CRITICAL,"Link removal uncomplete in %s at line %d",
		 __FILE__,__LINE__);
    }

    dataBase->nbLinks--;

    return ret;
}

int getpreemptList(DataBase *dataBase,DBLabelSwitchedPath *lsp,LongList *preemptList)
{
    DBLSPList *lspList;
    int i;
    DBLink *lnk=NULL;
    LongVec isProcessed; 
    double rerouteGain[NB_OA];
    bool allowLSP=TRUE;
#if defined SIMULATOR
    LongList idList;

    if (dataBase == NULL || lsp==NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }


    if (longVecInit(&(isProcessed), dataBase->linkSrcVec.size)<0)
    {
	addError(CRITICAL,"Unable to initialize vector of longs in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    memset(rerouteGain,0,NB_OA*sizeof(double));

    // Check if establishment is possible

    if (longListInit(&(idList),-1)<0)
    {
	addError(CRITICAL,"Unable to initialize list of longs in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    for (i=0;(i<lsp->path.top-1) && allowLSP;i++)
    {
	lnk=DBlinkTabGet(&(dataBase->linkTab),lsp->path.cont[i],
			 lsp->path.cont[i+1]);

	

	allowLSP= allowLSP && isValidLSPLink(dataBase,lsp->path.cont[i],lsp->path.cont[i+1],
					     &(lnk->state),lsp,rerouteGain);
	if ((lspList=DBgetLinkLSPs(dataBase,lsp->path.cont[i],lsp->path.cont[i+1]))==NULL)
	{
	    addError(CRITICAL,"Unable to get the list of LSPs carried by the link in %s at line %d",
		     __FILE__,__LINE__);
	    longListEnd(&(idList));
	    longVecEnd(&(isProcessed));
	    return -1;
	}
	idList.top=0;
	if (chooseReroutedLSPs(lsp->precedence,&(lnk->state),lspList,rerouteGain,&(idList))<0)
	{
	    addError(CRITICAL,"Unable choose LSPs for rerouting in %s at line %d",
		     __FILE__,__LINE__);
	    longListEnd(&(idList));
	    longVecEnd(&(isProcessed));
	    return -1;
	}
	if (longListMerge(&(idList),preemptList,preemptList)<0)
	{
	    addError(CRITICAL,"Unable to merge lists of longs in %s at line %d",
		     __FILE__,__LINE__);
	    longListEnd(&(idList));
	    longVecEnd(&(isProcessed));
	    return -1;
	}
	isProcessed.cont[lnk->id] = 1;
    }
    if ((lsp->type == GLOBAL_BACK) || (lsp->type == LOCAL_BACK))
    {
	for (i=0;(i<lsp->primPath.top-1) && allowLSP;i++)
	{
	    lnk=DBlinkTabGet(&(dataBase->linkTab),lsp->primPath.cont[i],
			     lsp->primPath.cont[i+1]);
	    if (isProcessed.cont[lnk->id] == 0)
	    {
		allowLSP= allowLSP && isValidLSPLink(dataBase,lsp->primPath.cont[i],lsp->primPath.cont[i+1],
						 &(lnk->state),lsp,rerouteGain);
		if ((lspList=DBgetLinkLSPs(dataBase,lsp->primPath.cont[i],lsp->primPath.cont[i+1]))==NULL)
		{
		    addError(CRITICAL,"Unable to get the list of LSPs carried by the link in %s at line %d",
			     __FILE__,__LINE__);
		    longListEnd(&(idList));
		    longVecEnd(&(isProcessed));
		    return -1;
		}
		idList.top=0;
		if (chooseReroutedLSPs(lsp->precedence,&(lnk->state),lspList,rerouteGain,&(idList))<0)
		{
		    addError(CRITICAL,"Unable choose LSPs for rerouting in %s at line %d",
			     __FILE__,__LINE__);
		    longListEnd(&(idList));
		    longVecEnd(&(isProcessed));
		    return -1;
		}
		if (longListMerge(&(idList),preemptList,preemptList)<0)
		{
		    addError(CRITICAL,"Unable to merge lists of longs in %s at line %d",
			     __FILE__,__LINE__);
		    longListEnd(&(idList));
		    longVecEnd(&(isProcessed));
		    return -1;
		}
		isProcessed.cont[lnk->id] = 1;
	    }
	}
    }
    longListEnd(&(idList));
    return 0;

#elif defined AGENT 
    addError(CRITICAL,"The function getpreemptList is not supposed to be called in AGENT mode  in %s at line %d",
	     __FILE__,__LINE__);
#endif 
    return -1;
}

int DBaddLSP(DataBase *dataBase,DBLabelSwitchedPath *lsp,LongList *preemptList)
{
    DBLabelSwitchedPath *newLSP=NULL, *contentLSP=NULL;
    DBLSPList *lspList;
    int i,ret=0;
    DBLink *lnk=NULL;
    LongVec isProcessed; 
    double rerouteGain[NB_OA];
    bool allowLSP=TRUE;
#if defined SIMULATOR
    LongList idList;
#elif defined AGENT
    int j;
    bool inPath=FALSE;
#endif

#if defined LINUX && defined TIME2
    struct timezone tz;
    struct timeval  t1,t2;
#endif

    if (dataBase == NULL || lsp==NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    

#if defined LINUX && defined TIME2
    gettimeofday(&t1, &tz);
#endif

    if (longVecInit(&(isProcessed), dataBase->linkSrcVec.size)<0)
    {
	addError(CRITICAL,"Unable to initialize vector of longs in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    memset(rerouteGain,0,NB_OA*sizeof(double));

    // Check if establishment is possible
#if defined SIMULATOR
    if (longListInit(&(idList),-1)<0)
    {
	addError(CRITICAL,"Unable to initialize list of longs in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }
    
    for (i=0;(i<lsp->path.top-1) && allowLSP;i++)
    {
	
	lnk=DBlinkTabGet(&(dataBase->linkTab),lsp->path.cont[i],
			 lsp->path.cont[i+1]);
	/*if (lnk == NULL){
	    abort();
	    }*/
	allowLSP= allowLSP && isValidLSPLink(dataBase,lsp->path.cont[i],lsp->path.cont[i+1],
					     &(lnk->state),lsp,rerouteGain);
	if ((lspList=DBgetLinkLSPs(dataBase,lsp->path.cont[i],lsp->path.cont[i+1]))==NULL)
	{
	    addError(CRITICAL,"Unable to get the list of LSPs carried by the link in %s at line %d",
		     __FILE__,__LINE__);
	    longListEnd(&(idList));
	    longVecEnd(&(isProcessed));
	    return -1;
	}
	idList.top=0;
	if (chooseReroutedLSPs(lsp->precedence,&(lnk->state),lspList,rerouteGain,&(idList))<0)
	{
	    addError(CRITICAL,"Unable choose LSPs for rerouting in %s at line %d",
		     __FILE__,__LINE__);
	    longListEnd(&(idList));
	    longVecEnd(&(isProcessed));
	    return -1;
	}
	if (longListMerge(&(idList),preemptList,preemptList)<0)
	{
	    addError(CRITICAL,"Unable to merge lists of longs in %s at line %d",
		     __FILE__,__LINE__);
	    longListEnd(&(idList));
	    longVecEnd(&(isProcessed));
	    return -1;
	}
	isProcessed.cont[lnk->id] = 1;
    }
    
    if ((lsp->type == GLOBAL_BACK) || (lsp->type == LOCAL_BACK))
    {
	for (i=0;(i<lsp->primPath.top-1) && allowLSP;i++)
	{
	    lnk=DBlinkTabGet(&(dataBase->linkTab),lsp->primPath.cont[i],
			     lsp->primPath.cont[i+1]);
	    if (isProcessed.cont[lnk->id] == 0)
	    {
		allowLSP= allowLSP && isValidLSPLink(dataBase,lsp->primPath.cont[i],lsp->primPath.cont[i+1],
						 &(lnk->state),lsp,rerouteGain);
		if ((lspList=DBgetLinkLSPs(dataBase,lsp->primPath.cont[i],lsp->primPath.cont[i+1]))==NULL)
		{
		    addError(CRITICAL,"Unable to get the list of LSPs carried by the link in %s at line %d",
			     __FILE__,__LINE__);
		    longListEnd(&(idList));
		    longVecEnd(&(isProcessed));
		    return -1;
		}
		idList.top=0;
		if (chooseReroutedLSPs(lsp->precedence,&(lnk->state),lspList,rerouteGain,&(idList))<0)
		{
		    addError(CRITICAL,"Unable choose LSPs for rerouting in %s at line %d",
			     __FILE__,__LINE__);
		    longListEnd(&(idList));
		    longVecEnd(&(isProcessed));
		    return -1;
		}
		if (longListMerge(&(idList),preemptList,preemptList)<0)
		{
		    addError(CRITICAL,"Unable to merge lists of longs in %s at line %d",
			     __FILE__,__LINE__);
		    longListEnd(&(idList));
		    longVecEnd(&(isProcessed));
		    return -1;
		}
		isProcessed.cont[lnk->id] = 1;
	    }
	}
    }
    longListEnd(&(idList));
#elif defined AGENT
    for (i=0;(i<lsp->path.top-1) && (lsp->path.cont[i]!=dataBase->id);i++);

    if (i<lsp->path.top-1)
    {
	lnk=DBlinkTabGet(&(dataBase->linkTab),lsp->path.cont[i],
			 lsp->path.cont[i+1]);
	allowLSP= allowLSP && isValidLSPLink(dataBase,lsp->path.cont[i],lsp->path.cont[i+1],
					     &(lnk->state),lsp,rerouteGain);
	if ((lspList=DBgetLinkLSPs(dataBase,lsp->path.cont[i],lsp->path.cont[i+1]))==NULL)
	{
	    addError(CRITICAL,"Unable to get the list of LSPs carried by the link in %s at line %d",
		     __FILE__,__LINE__);
	    longVecEnd(&(isProcessed));
	}
	if (chooseReroutedLSPs(lsp->precedence,&(lnk->state),lspList,rerouteGain,preemptList)<0)
	{
	    addError(CRITICAL,"Unable choose LSPs for rerouting in %s at line %d",
		     __FILE__,__LINE__);
	    longVecEnd(&(isProcessed));
	    return -1;
	}
	isProcessed.cont[lnk->id] = 1;
	inPath=TRUE;
    }
    if ((lsp->type == GLOBAL_BACK) || (lsp->type == LOCAL_BACK))
    {
	for (j=0;(j<lsp->primPath.top-1) && (lsp->primPath.cont[j]!=dataBase->id);j++);

	if (j<lsp->primPath.top-1)
	{
	    lnk=DBlinkTabGet(&(dataBase->linkTab),lsp->primPath.cont[j],
			     lsp->primPath.cont[j+1]);
	    if (isProcessed.cont[lnk->id] == 0)
	    {
		allowLSP= allowLSP && isValidLSPLink(dataBase,lsp->primPath.cont[j],lsp->primPath.cont[j+1],
						 &(lnk->state),lsp,rerouteGain);
		if ((lspList=DBgetLinkLSPs(dataBase,lsp->primPath.cont[i],lsp->primPath.cont[i+1]))==NULL)
		{
		    addError(CRITICAL,"Unable to get the list of LSPs carried by the link in %s at line %d",
			     __FILE__,__LINE__);
		    longVecEnd(&(isProcessed));
		}
		if (chooseReroutedLSPs(lsp->precedence,&(lnk->state),lspList,rerouteGain,preemptList)<0)
		{
		    addError(CRITICAL,"Unable choose LSPs for rerouting in %s at line %d",
			     __FILE__,__LINE__);
		    longVecEnd(&(isProcessed));
		    return -1;
		}
		isProcessed.cont[lnk->id] = 1;
	    }
	    inPath=TRUE;
	}
    }
    if (!inPath)
    {
	addError(CRITICAL,"Agent not concerned by this LSP in %s at line %d",
		 __FILE__,__LINE__);
	longVecEnd(&(isProcessed));
	return -1;
    }
#else
    // Generate an error;
    COMPILE_ERROR;
#endif

    if (!allowLSP)
    {
	addError(CRITICAL,"LSP refused by the predicate in %s at line %d",
		 __FILE__,__LINE__);
	longVecEnd(&(isProcessed));
	return -1;
    }


    if ((newLSP=DBlspNew())==NULL)
    {
	addError(CRITICAL,"Unable to create LSP in %s at line %d",
		 __FILE__,__LINE__);
	longVecEnd(&(isProcessed));
	return -1;
    }
    
    if (DBlspCopy(newLSP,lsp)<0)
    {
	addError(CRITICAL,"Unable to create a valid LSP copy in %s at line %d",
		 __FILE__,__LINE__);
	DBlspDestroy(newLSP);
	longVecEnd(&(isProcessed));
	return -1;
    }	

    if (DBlspVecSet(&(dataBase->lspVec),newLSP,newLSP->id)<0)
    {
	addError(CRITICAL,"Unable to insert LSP in the general LSP container in %s at line %d",
		 __FILE__,__LINE__);
	DBlspDestroy(newLSP);
	longVecEnd(&(isProcessed));
	return -1;
    }

    if (newLSP->noContentionId>=0)
    {
	if ((contentLSP=DBlspVecGet(&(dataBase->lspVec),newLSP->noContentionId))==NULL)
	{
	    addError(WARNING,"Unable to get no contention LSP in %s at line %d",
		     __FILE__,__LINE__);
	    newLSP->noContentionId=-1;
	    // not critical enough to abort
	}
	else
	{
	    contentLSP->noContentionId=newLSP->id;
	}
    }

    for (i=0;i<isProcessed.size;i++)
    {
	isProcessed.cont[i]=0;	
    }


#if defined SIMULATOR
    // Add the LSP to each link list and update all the linkstates (only once !!!!!)
    
    for (i=0;i<newLSP->path.top-1;i++)
    {
	
	lnk=DBlinkTabGet(&(dataBase->linkTab),newLSP->path.cont[i],
			 newLSP->path.cont[i+1]);
	ANDERROR(ret,DBlspListInsert(&(lnk->lspList),newLSP)); 
	ANDERROR(ret,DBupdateLSOnSetup(dataBase, newLSP->path.cont[i], 
				       newLSP->path.cont[i+1], &(lnk->state), newLSP));
        
	isProcessed.cont[lnk->id] = 1;
	

    }
    if ((newLSP->type == GLOBAL_BACK) || (newLSP->type == LOCAL_BACK))
    {
	for (i=0;i<newLSP->primPath.top-1;i++)
	{
	    lnk=DBlinkTabGet(&(dataBase->linkTab),newLSP->primPath.cont[i],
			     newLSP->primPath.cont[i+1]);
	    if (isProcessed.cont[lnk->id] == 0)
	    {
		ANDERROR(ret,DBupdateLSOnSetup(dataBase, newLSP->primPath.cont[i], 
					       newLSP->primPath.cont[i+1], &(lnk->state), newLSP));
		isProcessed.cont[lnk->id] = 1;
	
		
	    }
	}
    }

#elif defined AGENT 
    // Add the LSP to the link attached to the agent and update the linkstate
    for (i=0;i<newLSP->path.top-1;i++)
    {
	lnk=DBlinkTabGet(&(dataBase->linkTab),newLSP->path.cont[i],
			 newLSP->path.cont[i+1]);
	ANDERROR(ret,DBlspListInsert(&(lnk->lspList),newLSP)); 

	if (newLSP->path.cont[i] == dataBase->id)
	{
	    ANDERROR(ret,DBupdateLSOnSetup(dataBase, newLSP->path.cont[i], 
					   newLSP->path.cont[i+1], &(lnk->state), newLSP));
	    isProcessed.cont[lnk->id] = 1;
	}
    }
    if ((newLSP->type == GLOBAL_BACK) || (newLSP->type == LOCAL_BACK))
    {
	for (i=0;i<newLSP->primPath.top-1;i++)
	{
	    lnk=DBlinkTabGet(&(dataBase->linkTab),newLSP->primPath.cont[i],
			     newLSP->primPath.cont[i+1]);

	    if (newLSP->primPath.cont[i] == dataBase->id)
	    {
		if (isProcessed.cont[lnk->id] == 0)
		{
		    ANDERROR(ret,DBupdateLSOnSetup(dataBase, newLSP->primPath.cont[i], 
						   newLSP->primPath.cont[i+1], &(lnk->state), newLSP));
		}
		break;
	    }
	}
    }
#else
    // Generate an error;
    COMPILE_ERROR;
#endif

    longVecEnd(&(isProcessed));

#if defined LINUX && defined TIME2
    gettimeofday(&t2, &tz);
    fprintf(stderr, "Time to add a new LSP : %f ms\n", (t2.tv_sec - t1.tv_sec) * 1000 + 
	    (t2.tv_usec - t1.tv_usec) / 1000.0);
#endif


    if (ret<0)
    {
	addError(CRITICAL,"LSP addition uncomplete in %s at line %d",
		 __FILE__,__LINE__);
    }
    
    return ret;
}

int DBremoveLSP(DataBase *dataBase, long id)
{
    DBLabelSwitchedPath *lsp=NULL, *contentLSP=NULL;
    int i,ret=0;
    DBLink *lnk=NULL;
    LongVec isProcessed;

    if (dataBase == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if ((lsp = DBlspVecGet(&(dataBase->lspVec), id)) == NULL)
    {
	addError(CRITICAL,"Trying to remove inexistent LSP (id = %ld) in %s at line %d",
		 id,__FILE__,__LINE__);
	return -1;
    }
    
    if (longVecInit(&(isProcessed), dataBase->linkSrcVec.size)<0)
    {
	addError(CRITICAL,"LSP removal uncomplete in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

#if defined SIMULATOR
    // Remove the LSP from each link list and update all the linkstates
    for (i=0;i<lsp->path.top-1;i++)
    {
	lnk=DBlinkTabGet(&(dataBase->linkTab),lsp->path.cont[i],
			 lsp->path.cont[i+1]);
	ANDERROR(ret,DBlspListRemove(&(lnk->lspList),lsp)); 
	ANDERROR(ret,DBupdateLSOnRemove(dataBase, lsp->path.cont[i], 
				       lsp->path.cont[i+1], &(lnk->state), lsp));
	isProcessed.cont[lnk->id] = 1;
    }
    if ((lsp->type == GLOBAL_BACK) || (lsp->type == LOCAL_BACK))
    {
	for (i=0;i<lsp->primPath.top-1;i++)
	{
	    lnk=DBlinkTabGet(&(dataBase->linkTab),lsp->primPath.cont[i],
			     lsp->primPath.cont[i+1]);
	    if (isProcessed.cont[lnk->id] == 0)
	    {
		ANDERROR(ret,DBupdateLSOnRemove(dataBase, lsp->primPath.cont[i], 
					       lsp->primPath.cont[i+1], &(lnk->state), lsp));
		isProcessed.cont[lnk->id] = 1;
	    }
	}
    }
#elif defined AGENT 
    // Remove the LSP to the link attached to the agent and update the linkstate
    for (i=0;i<lsp->path.top-1;i++)
    {
	lnk=DBlinkTabGet(&(dataBase->linkTab),lsp->path.cont[i],
			 lsp->path.cont[i+1]);
	ANDERROR(ret,DBlspListRemove(&(lnk->lspList),lsp)); 

	if (lsp->path.cont[i] == dataBase->id)
	{
	    ANDERROR(ret,DBupdateLSOnRemove(dataBase, lsp->path.cont[i], 
					   lsp->path.cont[i+1], &(lnk->state), lsp));
	    isProcessed.cont[lnk->id] = 1;
	}
    }
    if ((lsp->type == GLOBAL_BACK) || (lsp->type == LOCAL_BACK))
    {
	for (i=0;i<lsp->primPath.top-1;i++)
	{
	    lnk=DBlinkTabGet(&(dataBase->linkTab),lsp->primPath.cont[i],
			     lsp->primPath.cont[i+1]);

	    if (lsp->primPath.cont[i] == dataBase->id)
	    {
		if (isProcessed.cont[lnk->id] == 0)
		{
		    ANDERROR(ret,DBupdateLSOnRemove(dataBase, lsp->primPath.cont[i], 
						   lsp->primPath.cont[i+1], &(lnk->state), lsp));
		}
		break;
	    }
	}
    }
#else
    // Generate an error;
    COMPILE_ERROR;
#endif    

    longVecEnd(&(isProcessed));

    // remove the lsp from the global list
    ANDERROR(ret,DBlspVecRemove(&(dataBase->lspVec), id));

    if (lsp->noContentionId>=0)
    {
	if ((contentLSP=DBlspVecGet(&(dataBase->lspVec),lsp->noContentionId))==NULL)
	{
	    addError(WARNING,"Unable to get no contention LSP in %s at line %d",
		     __FILE__,__LINE__);
	    // not critical enough to abort
	}
	contentLSP->noContentionId=-1;
    }

    // free the lsp
    DBlspDestroy(lsp);

    if (ret<0)
    {
	addError(CRITICAL,"LSP removal uncomplete in %s at line %d",
		 __FILE__,__LINE__);
    }
    
    return ret;
}


DBLabelSwitchedPath* DBgetLSP(DataBase *dataBase,long id)
{
    if (dataBase == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;
    }

    return DBlspVecGet(&(dataBase->lspVec), id);
}


DBLSPList* DBgetLinkLSPs(DataBase *dataBase,long src,long dst)
{
    DBLink *lnk=NULL;

    if (dataBase == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;
    }

    if ((lnk=DBlinkTabGet(&(dataBase->linkTab),src,dst)) == NULL)
    {
	addError(CRITICAL,"Inexistent Link (src = %ld, dst = %ld) in %s at line %d",
		 src,dst,__FILE__,__LINE__);
	return NULL;
    }

    return &(lnk->lspList);
}

DBLinkState* DBgetLinkState(DataBase *dataBase,long src,long dst)
{
    DBLink *lnk=NULL;

    if (dataBase == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;
    }

    if ((lnk=DBlinkTabGet(&(dataBase->linkTab),src,dst)) == NULL)
    {
	addError(CRITICAL,"Inexistent Link (src = %ld, dst = %ld) in %s at line %d",
		 src,dst,__FILE__,__LINE__);
	return NULL;
    }

    return &(lnk->state);
}

int DBsetLinkState(DataBase *dataBase,long src,long dst,DBLinkState* newLS)
{
    DBLink *lnk=NULL;

    if (dataBase == NULL || newLS == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return -1;
    }

    if ((lnk=DBlinkTabGet(&(dataBase->linkTab),src,dst)) == NULL)
    {
	addError(CRITICAL,"Inexistent Link (src = %ld, dst = %ld) in %s at line %d",
		 src,dst,__FILE__,__LINE__);
	return -1;
    }

    if (DBlinkStateCopy(&(lnk->state), newLS)<0)
    {
	addError(CRITICAL,"Impossible to set linkstate on link (src = %ld, dst = %ld) in %s at line %d",
		 src,dst,__FILE__,__LINE__);
	return -1;
    }

    return 0;
}


LongList* DBgetNodeInNeighb(DataBase *dataBase,long id) 
{
    DBNode *node=NULL;

    if (dataBase == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;
    }

    if ((node=DBnodeVecGet(&(dataBase->nodeVec), id)) == NULL)
    {
	addError(CRITICAL,"Node don't exist (id = %ld) in %s at line %d",
		 id,__FILE__,__LINE__);
	return NULL;
    }

    return (&(node->inNeighb));
}


LongList* DBgetNodeOutNeighb(DataBase *dataBase,long id)
{
    DBNode *node=NULL;

    if (dataBase == NULL) 
    { 
	addError(CRITICAL,"Bad argument (NULL) in %s at line %d",
		 __FILE__,__LINE__);
	return NULL;
    }

    if ((node=DBnodeVecGet(&(dataBase->nodeVec), id)) == NULL)
    {
	addError(CRITICAL,"Node don't exist (id = %ld) in %s at line %d",
		 id,__FILE__,__LINE__);
	return NULL;
    }

    return (&(node->outNeighb));
}


void DBprintDB(DataBase* db)
{
    long i,j;

    printf("Printing info about nodes ...\n");
    printf("-----------------------------\n");

    for (i=0; i<db->nodeVec.size; i++) 
    {
	if (db->nodeVec.cont[i])
	{
	    printf("Node id : %ld\n", i);
	    printf("-------------\n");
	    DBprintNode(db->nodeVec.cont[i]);
	}
    }

    printf("\nPrinting info about links ...\n");
    printf("-----------------------------\n");

    for (i=0; i<db->linkTab.size; i++)
	for (j=0; j<db->linkTab.size; j++)
	{
	    if (db->linkTab.cont[i][j])
	    {
		printf("Link %ld-%ld (id = %ld)\n", i, j, DBgetLinkID(db, i, j));
		printf("-----------------------\n");

		DBprintLink(db->linkTab.cont[i][j]);

	    }
	}
}

