#include "database/database_api.h"
#include "computation/computation_api.h"
#include "computation/backup/backup_api.h"
#include "common/common.h"
#include "error/error.h"

#include <stdio.h>
#include <string.h>

//#define LINUX
//#define TIME1
//#define TIMING

#if defined LINUX && defined TIMING
#include <sys/time.h>
#endif

// Prototype for scenario :
extern int makeTopology(DataBase*);
extern int makeReqList(LSPRequestList*);

int
main() 
{
    DataBase* centralDB=NULL;
    LSPRequestList reqList,req;
    LongVec rerouting;
    DBLabelSwitchedPath lsp;
    long i,j;
    long id=0;

#if defined LINUX && defined TIMING && defined TIME1
    struct timezone tz;
    struct timeval  t1,t2;
#endif

    // Init the error manager ....
    errorInit();

    // Allocate a new DB
    if ((centralDB = DBnew(0)) == NULL)
    {
	addError(CRITICAL,"Impossible to init a DataBase in %s at line %d",
		 __FILE__,__LINE__);
	goto clean;
    }

    // Configuration
    damoteConfig.predicateConfig.colorClause=FALSE;
    damoteConfig.predicateConfig.allowReroute=FALSE;
    damoteConfig.predicateConfig.capacityClause=TRUE;
    
    memset(&damoteConfig.primaryComputationConfig,0,sizeof(PrimaryComputationConfig));
    damoteConfig.primaryComputationConfig.load[0]=1;
    
    memset(&damoteConfig.reroutingConfig,0,sizeof(ReroutingConfig));
    for (i=0;i<NB_OA;i++)
	for (j=0;j<NB_PREEMPTION;j++)
	    damoteConfig.reroutingConfig.scoreCoef[i][j]=1;
      
    makeTopology(centralDB);

    // Check the allocated mem
    printf("\n-----------------------------------------------------------------------\n");
    printf("Memory allocated on the heap : %ld bytes\n", allocatedMemory);

    // Create the list of request
    printf("\n-----------------------------------------------------------------------\n");
    printf("Preparing request list\n");
    lspRequestListInit(&reqList, -1);
    makeReqList(&reqList);

    // Check the allocated mem
    printf("\n-----------------------------------------------------------------------\n");
    printf("Memory allocated on the heap : %ld bytes\n", allocatedMemory);

    // init the lsp
    DBlspInit(&lsp);

    // Start establishement

#if defined LINUX && defined TIMING && defined TIME1
    gettimeofday(&t1, &tz);
#endif

    printf("Nb REQ = %ld\n", reqList.size);

    for (i=0; i<reqList.size; i++) {

	long j;

	if (i%10 == 0)
	    fprintf(stderr,"%ld already done\n", i);

	// compute
	if (computePrimaryPath(centralDB, &(reqList.cont[i])) >= 0) {

	    // print
/*
	    for (j=0; j<reqList.cont[i].path.top; j++) {
		printf("%ld ", reqList.cont[i].path.cont[j]);
	    }
	    printf("\n");
*/
	    // make the lsp
	    
	    longVecCopy(&(lsp.path), &(reqList.cont[i].path));
	    lsp.id = id++;
	    lsp.bw[0] = reqList.cont[i].bw[0];
	    lsp.precedence = 0;
	    lsp.type = PRIM;
	    lsp.primID = -1;
	    lsp.noContentionId = -1;

	    printf("Add LSP with id = %ld\n", lsp.id);

	    longVecInit(&rerouting,-1);
	    if (DBaddLSP(centralDB, &lsp, &rerouting) < 0)
	    {
		printf("[DBadd] Failure of LSP number %ld\n", i);
		id--;
	    }
	    else {
		long primID = lsp.id;	

		longVecEnd(&rerouting);

		lspRequestListInit(&req, -1);
		if (computeBackup(centralDB, &req, primID, LOCAL) < 0) {
		    printf("[Backup] Failure of LSP number %ld\n", i);
		    if (DBremoveLSP(centralDB,--id) < 0) {
			printf("[Backup] Failure of LSP removal %ld\n", id);
		    }
		}
		else {

		    longVecCopy(&(lsp.primPath), &(lsp.path));

		    for (j=0; j<req.size; j++) {

			// make the lsp
			printf("Valeur de req.size %ld\n", req.size);

			longVecCopy(&(lsp.path), &(req.cont[j].path));
			lsp.id = id++;
			lsp.bw[0] = req.cont[j].bw[0];
			lsp.precedence = 1;
			lsp.type = LOCAL_BACK;
			lsp.primID = primID;
			lsp.noContentionId = -1;

			printf("Add LSP with id = %ld, primID = %ld\n", lsp.id, primID);
			int u;
			for (u=0;u<lsp.path.top;u++)
			    printf(" %ld", lsp.path.cont[u]);
			
			printf(" \n");
				    
			longVecInit(&rerouting,-1);
			if (DBaddLSP(centralDB, &lsp, &rerouting) < 0)
			{
			    printf("[DBadd-BACKUP] Failure of LSP number %ld\n", lsp.id);
			    id--;
			    break;
			}
			longVecEnd(&rerouting);
		    }

		    if (j!=req.size) {
			// clean up
			for (j--; j>=0; j--) {
			    if (DBremoveLSP(centralDB,--id) < 0) {
				printf("[DBadd-Backup-CleanUP] Failure of LSP removal %ld\n", id);
			    }
			}
			// remove also the primary
			if (DBremoveLSP(centralDB,--id) < 0) {
			    printf("[DBadd-Backup-CleanUP] Failure of LSP removal %ld\n", id);
			}
		    }
		}
		lspRequestListEnd(&req);
	    }

	}
	else {
	    printf("[Computation] Failure of LSP number %ld\n", i);
	}
    }

#if defined LINUX && defined TIMING && defined TIME1
    gettimeofday(&t2, &tz);
    fprintf(stderr, "Time to establish all : %f ms\n", (t2.tv_sec - t1.tv_sec) * 1000 + 
	    (t2.tv_usec - t1.tv_usec) / 1000.0);
#endif

    // Print the DATABASE info
    DBprintDB(centralDB);
    
    // ------------------- END TEST 1 -----------------------
    
 clean:

    // Destroy the request list
    lspRequestListEnd(&reqList);
    
    // Destroy the lsp
    DBlspEnd(&lsp);

    // Destroy the DB if required
    if (centralDB != NULL)
	DBdestroy(centralDB);
    
    // Print list of error
    printErrorStack();
    
    // Destroy error manager
    errorDestroy();
    
    // Check the allocated mem
    printf("\n-----------------------------------------------------------------------\n");
    printf("Memory still allocated on the heap : %ld bytes\n", allocatedMemory);
    
    return 0;
}
